@@ -596,7 +596,7 @@ module Make(X : ALIEN) = struct
 	    let unif_slic = equations_slice parts in
 	    let eq_pr = equalities_propagation unif_slic in
 	    let sol = build_solution unif_slic eq_pr in 
-	    if Options.debug_bitv then 
+	    if Options.debug_bitv () then 
 	      begin
 		Print.print_sliced_sys err st_sys;
 		Print.print_c_solve_res err sys_sols;
@@ -704,12 +704,12 @@ module Make(X : ALIEN) = struct
 
 
   (* ne resout pas quand c'est deja resolu *)
-  let solve u t = 
-    if Options.debug_bitv then 
+  let solve_bis u t = 
+    if Options.debug_bitv () then 
       eprintf "[Bitv] solve %a = %a@." X.print u X.print t;
     
     match X.extract u , X.extract t with 
-      | None   , None   -> assert false
+      | None   , None   -> if X.compare u t > 0 then [u,t] else [t,u]
       | None   , Some _ -> [u , t]
       | Some _ , None   -> [t , u]
       | Some u , Some t -> 
@@ -737,7 +737,7 @@ module Make(X : ALIEN) = struct
 	      bv = Canonizer.I_Comp(subst_rec x subs u ,subst_rec x subs v)}
 
   let subst x subs biv = 
-    if Options.debug_bitv then 
+    if Options.debug_bitv () then 
       eprintf "[Bitv] subst %a |-> %a in %a@." X.print x X.print subs print biv;
     if biv = [] then is_mine biv
     else 
@@ -767,17 +767,24 @@ module Make(X : ALIEN) = struct
 
   let term_extract _ = None
 
+  let abstract_selectors v acc = is_mine v, acc
+
+  let solve r1 r2 pb = 
+    {pb with sbt = List.rev_append (solve_bis r1 r2) pb.sbt}
+
   module Rel = struct
 
     type r = X.r
     type t =  unit
 
-    let empty () = ()
-    let assume _ _ ~are_eq ~are_neq ~class_of =
+    let empty _ = ()
+    let assume _ _ ~are_eq ~are_neq ~class_of ~classes =
       (), { assume = []; remove = [] }
     let add _ _ = ()
     let case_split _ = []    
-    let query _ _ ~are_eq ~are_neq ~class_of = Sig.No
+    let query _ _ ~are_eq ~are_neq ~class_of ~classes = Sig.No
+    let print_model _ _ _ = ()
+    let new_terms env = Term.Set.empty
   end
 
 end
