@@ -10,6 +10,7 @@
 (*     Mohamed Iguernelala                                                *)
 (*     Stephane Lescuyer                                                  *)
 (*     Alain Mebsout                                                      *)
+(*     Claire Dross                                                       *)
 (*                                                                        *)
 (*     CNRS - INRIA - Universite Paris Sud                                *)
 (*                                                                        *)
@@ -20,3 +21,9 @@
 open Why_ptree
 
 val make : ((int tdecl, int) annoted * bool) list -> sat_tdecl Queue.t
+
+(* For the formulas of a theory. Simplify future handling by applying
+   DeMorgan rules. A formula is a conjuction of disjunctions of a literal
+   and a formula or a quantified formula. The order of the input is preserved.
+*)
+val make_theory : ((int tdecl, int) annoted * bool) list -> sat_tdecl Queue.t
