@@ -23,7 +23,8 @@ type lemma =
     { qvars: Symbols.Set.t;  (* toplevel quantified variables *)
       triggers : (Term.t list * Literal.LT.t option) list; (* multi-triggers *)
       main : t;  (* the main lemma's formula *)
-      name : string; }
+      name : string; 
+    }
       
 and llet = {
   let_var: Symbols.t;
@@ -53,8 +54,7 @@ val mk_if : Term.t -> t -> t -> int -> t
 val mk_iff : t -> t -> int -> t
 val mk_lit : Literal.LT.t -> int -> t
 val mk_forall : Term.Set.t -> Term.Set.t -> 
-  (Term.t list * Literal.LT.t option) list -> t ->
-  string -> int -> t
+  (Term.t list * Literal.LT.t option) list -> t -> string -> int -> t
 val mk_exists : Term.Set.t -> Term.Set.t -> 
   (Term.t list * Literal.LT.t option) list -> t ->
   string -> int -> t
@@ -62,6 +62,7 @@ val mk_let : Term.Set.t -> Symbols.t -> Term.t -> t -> int -> t
 
 val add_label : Hstring.t -> t -> unit
 val label : t -> Hstring.t
+val is_in_model : t -> bool
 
 val view : t -> view
 val size : t -> int
