@@ -10,6 +10,7 @@
 (*     Mohamed Iguernelala                                                *)
 (*     Stephane Lescuyer                                                  *)
 (*     Alain Mebsout                                                      *)
+(*     Claire Dross                                                       *)
 (*                                                                        *)
 (*     CNRS - INRIA - Universite Paris Sud                                *)
 (*                                                                        *)
@@ -21,13 +22,16 @@ open Why_ptree
 
 module Time : sig
 
-  val start: unit -> unit
+  val start : unit -> unit
+  val get : unit -> float
 
-  val get: unit -> float
+  val set_timeout : unit -> unit
+  val unset_timeout : unit -> unit
 
 end
 
-type output = Unsat of Explanation.t | Inconsistent | Sat | Unknown
+type output = Unsat of Explanation.t | Inconsistent 
+	      | Sat of Sat.t | Unknown of Sat.t
 
 
 val process_decl:
@@ -42,3 +46,8 @@ val open_file:
 val processing:
   (Why_ptree.sat_tdecl -> output -> int64 -> 'a) -> 
   ((int tdecl, int) annoted * Why_typing.env) list list -> unit
+
+
+val print_status : sat_tdecl -> output -> int64 -> unit
+
+val main : unit -> unit
