@@ -28,6 +28,12 @@ type borne =
   | Large of (num * Ex.t) 
   | Pinfty | Minfty
 
+
+let num_0 = Int 0
+let num_1 = Int 1
+let num_minus_1 = Int (-1)
+
+
 let compare_bornes b1 b2 =
   match b1, b2 with
     | Minfty, Minfty | Pinfty, Pinfty -> 0
@@ -39,18 +45,24 @@ let compare_bornes b1 b2 =
 
 let compare_bu_bl b1 b2 =
   match b1, b2 with
-    | (Minfty | Pinfty), _ | _,(Minfty | Pinfty) 
-    | Strict _, Strict _ | Large _, Large _ -> 
-      compare_bornes b1 b2 
+    | (Minfty | Pinfty), _ | _,(Minfty | Pinfty)
+    | Large _, Large _ -> 
+      compare_bornes b1 b2
+    | Strict _, Strict _ ->
+      let c = compare_bornes b1 b2 in
+      if c = 0 then -1 else c
     | Strict (v1, _), Large (v2, _) | Large (v1, _), Strict (v2, _) ->
       let c = compare_num v1 v2 in
       if c = 0 then -1 else c
       
 let compare_bl_bu b1 b2 =
   match b1, b2 with
-    | (Minfty | Pinfty), _ | _,(Minfty | Pinfty) 
-    | Strict _, Strict _ | Large _, Large _ -> 
-      compare_bornes b1 b2 
+    | (Minfty | Pinfty), _ | _,(Minfty | Pinfty)
+    | Large _, Large _ -> 
+      compare_bornes b1 b2
+    | Strict _, Strict _ ->
+      let c = compare_bornes b1 b2 in
+      if c = 0 then 1 else c
     | Strict (v1, _), Large (v2, _) | Large (v1, _), Strict (v2, _) ->
       let c = compare_num v1 v2 in
       if c = 0 then 1 else c
@@ -89,27 +101,37 @@ exception EmptyInterval of Ex.t
 exception NotConsistent of Ex.t
 exception Not_a_float
 
-let print_borne fmt = function
-  | Minfty -> fprintf fmt "-inf" 
-  | Pinfty -> fprintf fmt "+inf"
+let print_borne pretty fmt = function
+  | Minfty -> fprintf fmt "%s" (if pretty then "-∞" else "-inf")
+  | Pinfty -> fprintf fmt "%s" (if pretty then "+∞" else "+inf")
   | Strict (v, e) | Large (v, e) ->
-    if verbose || proof then 
+    if verbose () || proof () then 
       fprintf fmt "%s %a" (string_of_num v) Ex.print e
     else fprintf fmt "%s" (string_of_num v)
       
-let print_interval fmt (b1,b2) =
+let print_interval pretty fmt (b1,b2) =
   let c1, c2 = match b1, b2 with
     | Large _, Large _ -> '[', ']'
     | Large _, _ -> '[', '['
     | _, Large _ -> ']', ']'
     | _, _ -> ']', '['
   in 	    
-  fprintf fmt "%c%a;%a%c" c1 print_borne b1 print_borne b2 c2
-    
-let print fmt {ints = ints; is_int = b; expl = e } = 
-  List.iter (fun i -> fprintf fmt "%a" print_interval i) ints;
-  if verbose || proof then fprintf fmt " %a" Ex.print e
+  fprintf fmt "%c%a;%a%c" c1 (print_borne pretty) b1 (print_borne pretty) b2 c2
+
+let rec print_list pretty fmt = function
+  | [] -> ()
+  | [t] -> print_interval pretty fmt t
+  | t::l -> fprintf fmt "%a%s%a" 
+    (print_interval pretty) t 
+    (if pretty then " ∪ " else " ") (print_list pretty) l
+
+let pretty_print fmt {ints = ints; is_int = b; expl = e } = 
+  print_list true fmt ints;
+  if verbose () || proof () then fprintf fmt " %a" Ex.print e
   
+let print fmt {ints = ints; is_int = b; expl = e } = 
+  print_list false fmt ints;
+  if verbose () || proof () then fprintf fmt " %a" Ex.print e
 
 let undefined ty = {
   ints = [Minfty, Pinfty];
@@ -144,9 +166,9 @@ let is_point { ints = l; expl = e } =
 let add_expl_zero i expl =
   let res = List.map (fun x -> 
     match x with
-      | (Large ((Num.Int 0), e1) , Large ((Num.Int 0), e2)) ->
-        (Large ((Num.Int 0), Ex.union e1 expl),
-         Large ((Num.Int 0), Ex.union e2 expl))
+      | (Large (Int 0, e1) , Large (Int 0, e2)) ->
+        (Large (Int 0, Ex.union e1 expl),
+         Large (Int 0, Ex.union e2 expl))
       | _ -> x) i.ints in
   { i with ints = res }
 
@@ -191,15 +213,15 @@ let max_borne b1 b2 =
 	  | _, _ -> b1
 	
 let pos_borne b1 =
-  compare_bornes b1 (borne_of true Ex.empty (Int 0)) >= 0
+  compare_bornes b1 (borne_of true Ex.empty num_0) >= 0
 let pos_borne_strict b1 = 
-  compare_bornes b1 (borne_of true Ex.empty (Int 0)) > 0
+  compare_bornes b1 (borne_of true Ex.empty num_0) > 0
 let neg_borne b1 = 
-  compare_bornes b1 (borne_of true Ex.empty (Int 0)) <= 0
+  compare_bornes b1 (borne_of true Ex.empty num_0) <= 0
 let neg_borne_strict b1 = 
-  compare_bornes b1 (borne_of true Ex.empty (Int 0)) < 0
+  compare_bornes b1 (borne_of true Ex.empty num_0) < 0
 let zero_borne b1 = 
-  compare_bornes b1 (borne_of true Ex.empty (Int 0)) = 0
+  compare_bornes b1 (borne_of true Ex.empty num_0) = 0
 
 exception Found of Sig.answer
 
@@ -212,9 +234,9 @@ let doesnt_contain_0 {ints=l} =
 	  raise (Found 
 		   (Sig.Yes 
 		      (Ex.union 
-			 (explain_borne old_u) (explain_borne l))));
+			 (explain_borne old_u) (explain_borne l), [])));
 	u) Minfty l in
-    if neg_borne_strict max then Sig.Yes (explain_borne max)
+    if neg_borne_strict max then Sig.Yes (explain_borne max, [])
     else Sig.No
   with Found ans -> ans
 
@@ -225,13 +247,13 @@ let is_strict_smaller i1 i2 =
     | _ ->
       try
 	List.iter2 (fun (l1, u1) (l2, u2) ->
-	  if compare_bornes l1 l2 > 0 || compare_bornes u1 u2 < 0
+	  if compare_bl_bl l1 l2 > 0 || compare_bu_bu u1 u2 < 0
 	  then raise Exit
 	) i1 i2;
 	false
       with 
 	| Exit -> true
-	| Invalid_argument _ -> List.length i1 > List.length i2
+	| Invalid_argument _ -> true
 
 let is_strict_smaller {ints=i1} {ints=i2} = 
   is_strict_smaller i1 i2
@@ -241,9 +263,9 @@ let rec union_bornes l =
   match l with
     | [] | [_] -> l
     | (l1, u1)::((l2, u2)::r as r2) ->
-	if compare_bornes u1 l2 < 0 then
+	if compare_bu_bl u1 l2 < 0 then
 	  (l1, u1)::(union_bornes r2)
-	else if compare_bornes u1 u2 > 0 then
+	else if compare_bu_bu u1 u2 > 0 then
 	  union_bornes ((l1, u1)::r)
 	else
 	  union_bornes ((l1, u2)::r)
@@ -283,25 +305,25 @@ let minus_borne = function
   | Strict (v, e) -> Strict (minus_num v, e)
 
 let scale_borne n b =
-  assert (n >=/ Int 0);
-  if n =/ Int 0 then 
+  assert (n >=/ num_0);
+  if n =/ num_0 then 
     match b with
-    | Pinfty | Minfty -> Large (Int 0, Ex.empty)
-    | Large (_, e) | Strict (_, e) ->  Large (Int 0, e)
+    | Pinfty | Minfty -> Large (num_0, Ex.empty)
+    | Large (_, e) | Strict (_, e) ->  Large (num_0, e)
   else match b with
     | Pinfty | Minfty -> b
     | Large (v, e) -> Large (n */ v, e)
     | Strict (v, e) -> Strict (n */ v, e)
 
 let scale_interval n (b1,b2) =
-  if n </ Int 0 then
+  if n </ num_0 then
     (minus_borne (scale_borne (minus_num n) b2),
      minus_borne (scale_borne (minus_num n) b1))
   else (scale_borne n b1, scale_borne n b2)
 
 
 let scale n uints =
-  if rules = 4 then fprintf fmt "[rule] TR-Arith-Axiomes scale@.";
+  if rules () = 4 then fprintf fmt "[rule] TR-Arith-Axiomes scale@.";
   let l = List.map (scale_interval n) uints.ints in
   union { uints with ints = l; expl = uints.expl }
 	    
@@ -309,12 +331,12 @@ let mult_borne b1 b2 =
   match b1,b2 with
     | Minfty, Pinfty | Pinfty, Minfty -> assert false
     | Minfty, b | b, Minfty ->
-	if compare_bornes b (borne_of true Ex.empty (Int 0)) = 0 
+	if compare_bornes b (borne_of true Ex.empty num_0) = 0 
         then b
 	else if pos_borne b then Minfty
 	else Pinfty
     | Pinfty, b | b, Pinfty ->
-	if compare_bornes b (borne_of true Ex.empty (Int 0)) = 0 
+	if compare_bornes b (borne_of true Ex.empty num_0) = 0 
         then b
 	else if pos_borne b then Pinfty
 	else Minfty
@@ -397,7 +419,7 @@ let power_bornes p (b1,b2) =
       | p when p mod 2 = 0 ->
 	  (* max_merge to have explanations !!! *)
 	  let m = max_merge (power_borne_sup p b1) (power_borne_sup p b2) in
-	  (Large (Int 0, Ex.empty), m)
+	  (Large (num_0, Ex.empty), m)
       | _ -> (power_borne_inf p b1, power_borne_sup p b2)
   else if pos_borne b1 && pos_borne b2 then
     (power_borne_inf p b1, power_borne_sup p b2)
@@ -414,7 +436,7 @@ let int_of_borne_inf b =
     | Large (v, e) -> Large (ceiling_num v, e)
     | Strict (v, e) ->
 	let v' = ceiling_num v in
-	if v' >/ v then Large (v', e) else Large (v +/ (Int 1), e) 
+	if v' >/ v then Large (v', e) else Large (v +/ num_1, e) 
 
 let int_of_borne_sup b =
   match b with
@@ -422,7 +444,7 @@ let int_of_borne_sup b =
     | Large (v, e) -> Large (floor_num v, e)
     | Strict (v, e) ->
 	let v' = floor_num v in
-	if v' </ v then Large (v', e) else Large (v -/ (Int 1), e) 
+	if v' </ v then Large (v', e) else Large (v -/ num_1, e) 
 
 let int_div_of_borne_inf b =
   match b with
@@ -430,7 +452,7 @@ let int_div_of_borne_inf b =
     | Large (v, e) -> Large (floor_num v, e)
     | Strict (v, e) ->
 	let v' = floor_num v in
-	if v' >/ v then Large (v', e) else Large (v +/ (Int 1), e) 
+	if v' >/ v then Large (v', e) else Large (v +/ num_1, e) 
 
 let int_div_of_borne_sup b =
   match b with
@@ -438,7 +460,7 @@ let int_div_of_borne_sup b =
     | Large (v, e) -> Large (floor_num v, e)
     | Strict (v, e) ->
 	let v' = floor_num v in
-	if v' </ v then Large (v', e) else Large (v -/ (Int 1), e)
+	if v' </ v then Large (v', e) else Large (v -/ num_1, e)
 
 let int_bornes l u = 
   int_of_borne_inf l, int_of_borne_sup u
@@ -532,7 +554,7 @@ let exclude uints1 uints2 =
   intersect (complement uints1) uints2 
 
 let mult u1 u2 =
-  if rules = 4 then fprintf fmt "[rule] TR-Arith-Axiomes mult@.";
+  if rules () = 4 then fprintf fmt "[rule] TR-Arith-Axiomes mult@.";
   let resl, expl = 
     List.fold_left
       (fun (l', expl) b1 ->
@@ -547,7 +569,7 @@ let mult u1 u2 =
                  (Ex.union u1.expl u2.expl) }
 
 let power n u =
-  if rules = 4 then fprintf fmt "[rule] TR-Arith-Axiomes power@.";
+  if rules () = 4 then fprintf fmt "[rule] TR-Arith-Axiomes power@.";
   let l = List.map (power_bornes n) u.ints in
   union { u with ints = l }
 
@@ -568,33 +590,39 @@ let num_of_float x =
   (Big_int z) */ factor
 
 let root_num a n = 
-  if a </ (Int 0) then assert false
-  else if a =/ (Int 0) then (Int 0)
-  else if n = 2 then num_of_float (sqrt (float_of_num a))
-  else num_of_float ((float_of_num a) ** (1./. (float n)))
+  if a </ num_0 then assert false
+  else if a =/ num_0 then num_0
+  else
+    let v = float_of_num a in
+    let w = if v < min_float then min_float
+      else if v > max_float then max_float
+      else v
+    in
+    if n = 2 then num_of_float (sqrt w)
+    else num_of_float (w ** (1./. (float n)))
 
 let root_default_num a n =
   let s = root_num a n in
   let d = a -/ (s **/ (Int n)) in
-  if d >=/ (Int 0) then s else a // (s **/ ((Int n) -/ (Int 1)))
+  if d >=/ num_0 then s else a // (s **/ ((Int n) -/ num_1))
 
 let root_exces_num a n =
   let s = root_num a n in
   let d = a -/ (s **/ (Int n)) in
-  if d <=/ (Int 0) then s else a // (s **/ ((Int n) -/ (Int 1)))
+  if d <=/ num_0 then s else a // (s **/ ((Int n) -/ num_1))
 
 let root_default_borne is_int x n =
   match x with
     | Pinfty -> Pinfty
     | Minfty -> Minfty
     | Large (v, e) | Strict (v, e) ->
-	let s = if v >=/ (Int 0) then root_default_num v n
+	let s = if v >=/ num_0 then root_default_num v n
 	else (minus_num (root_exces_num (minus_num v) n)) in
 	if is_int then
 	  let cs = ceiling_num s in
 	  let cs2 = cs **/ (Int n) in
 	  if v <=/ cs2 then Large (cs, e)
-	  else Large (cs +/ (Int 1), e)
+	  else Large (cs +/ num_1, e)
 	else Large (s, e)
 
 let root_exces_borne is_int x n =
@@ -602,13 +630,13 @@ let root_exces_borne is_int x n =
     | Pinfty -> Pinfty
     | Minfty -> Minfty
     | Large (v, e) | Strict (v, e) ->
-	let s = if v >=/ (Int 0) then root_exces_num v n
+	let s = if v >=/ num_0 then root_exces_num v n
 	else (minus_num (root_default_num (minus_num v) n)) in
 	if is_int then
 	  let cs = floor_num s in
 	  let cs2 = cs **/ (Int n) in
 	  if v >=/ cs2 then Large (cs, e)
-	  else Large (cs -/ (Int 1), e)
+	  else Large (cs -/ num_1, e)
 	else Large (s, e)
 
 let sqrt_interval is_int (b1,b2) =
@@ -627,7 +655,7 @@ let root_interval is_int (b1,b2) n =
   if compare_bornes u l > 0 then [] else [u,l]
 
 let sqrt {ints = l; is_int = is_int; expl = e } =
-  if rules = 4 then fprintf fmt "[rule] TR-Arith-Axiomes sqrt@.";
+  if rules () = 4 then fprintf fmt "[rule] TR-Arith-Axiomes sqrt@.";
   let l =
     List.fold_left
       (fun l' bs ->
@@ -636,7 +664,7 @@ let sqrt {ints = l; is_int = is_int; expl = e } =
   union { ints = l; is_int = is_int; expl = e }
 
 let rec root n ({ints = l; is_int = is_int; expl = e} as u) =
-  if rules = 4 then fprintf fmt "[rule] TR-Arith-Axiomes root@.";
+  if rules () = 4 then fprintf fmt "[rule] TR-Arith-Axiomes root@.";
   if n mod 2 = 0 then root (n/2) (sqrt u)
   else
     let l =
@@ -655,9 +683,9 @@ let finite_size {ints = l; is_int = is_int} =
 	  (fun n (b1,b2) ->
 	     match b1, b2 with
 	       | Minfty, _ | _, Pinfty -> raise Exit
-	       | Large (v1, _) , Large (v2, _) -> n +/ (v2 -/ v1 +/ (Int 1))
+	       | Large (v1, _) , Large (v2, _) -> n +/ (v2 -/ v1 +/ num_1)
 	       | _, _ -> assert false
-	  ) (Int 0) l in
+	  ) num_0 l in
       Some n
     with Exit -> None
 		 
@@ -671,21 +699,21 @@ let inv_borne_inf b is_int ~other =
   match b with
     | Pinfty -> assert false
     | Minfty ->
-      if is_int then Large (Int 0,  explain_borne other) 
-      else Strict (Int 0, explain_borne other)
+      if is_int then Large (num_0,  explain_borne other) 
+      else Strict (num_0, explain_borne other)
     | Strict (Int 0, e) | Large (Int 0, e) -> Pinfty
-    | Strict (v, e) -> Strict (Int 1 // v, e)
-    | Large (v, e) -> Large (Int 1 // v, e)
+    | Strict (v, e) -> Strict (num_1 // v, e)
+    | Large (v, e) -> Large (num_1 // v, e)
 
 let inv_borne_sup b is_int ~other =
   match b with
     | Minfty -> assert false
     | Pinfty ->
-      if is_int then Large (Int 0, explain_borne other)
-      else Strict (Int 0, explain_borne other)
+      if is_int then Large (num_0, explain_borne other)
+      else Strict (num_0, explain_borne other)
     | Strict (Int 0, e) | Large (Int 0, e) -> Minfty
-    | Strict (v, e) -> Strict (Int 1 // v, e)
-    | Large (v, e) -> Large (Int 1 // v, e)
+    | Strict (v, e) -> Strict (num_1 // v, e)
+    | Large (v, e) -> Large (num_1 // v, e)
 
 let inv_bornes (l, u) is_int =
   inv_borne_sup u is_int ~other:l, inv_borne_inf l is_int ~other:u
@@ -704,12 +732,12 @@ let inv ({ints=l; is_int=is_int} as u) =
   with Exit -> { u with ints = [Minfty, Pinfty]  }
 
 let div i1 i2 =
-  if rules = 4 then fprintf fmt "[rule] TR-Arith-Axiomes div@.";
+  if rules () = 4 then fprintf fmt "[rule] TR-Arith-Axiomes div@.";
   let inv_i2 = inv i2 in
   if inv_i2.ints = [Minfty, Pinfty] then inv_i2
   else
     let i1 = match doesnt_contain_0 i2 with
-      | Sig.Yes ex -> add_expl_zero i1 ex
+      | Sig.Yes (ex, _) -> add_expl_zero i1 ex
       | Sig.No -> i1
     in
     let ({ints=l; is_int=is_int} as i) = mult i1 inv_i2 in
