@@ -35,8 +35,10 @@ type gformula = {
   f: F.t; 
   age: int; 
   name: F.t option; 
+  from_terms : Term.t list;
   mf: bool;
-  gf: bool
+  gf: bool;
+  inv : bool;
 }
 
 module H = Hashtbl.Make(Formula)
@@ -47,20 +49,22 @@ type t = {
     tbox : CcX.t;
     lemmas : (int * Ex.t) MF.t;
     definitions : (int * Ex.t) MF.t;
-    matching : MM.t
+    inversions :  (int * Ex.t) MF.t;
+    matching : MM.t;
+    add_inst: Formula.t -> unit;
 }
 
 exception Sat of t
 exception Unsat of Ex.t
-exception I_dont_know
-exception IUnsat of Ex.t
+exception I_dont_know of t
+exception IUnsat of Ex.t * Term.Set.t list
 
 let max_max_size = 96
 
 module Print = struct
 
-  let assume {f=f;age=age;name=lem;mf=mf} dep = 
-    if debug_sat then
+  let assume {f=f;age=age;name=lem;mf=mf;from_terms=terms} dep = 
+    if debug_sat () then
       begin
 	(match F.view f with
 	  | F.Unit _ -> ()
@@ -75,15 +79,16 @@ module Print = struct
 	      printf "@[%a@]@]@." F.print f
 		
 	  | F.Literal a -> 
-	      let s = 
-		match lem with 
-		    None -> "" 
-		  | Some ff -> 
-		      (match F.view ff with F.Lemma xx -> xx.F.name | _ -> "") 
+	      Term.print_list str_formatter terms;
+	      let s = flush_str_formatter () in
+	      let n = match lem with 
+		| None -> ""  
+		| Some ff -> 
+		    (match F.view ff with F.Lemma xx -> xx.F.name | _ -> "")
 	      in
 	      printf "\n@[@{<C.Bold>[sat]@}";
 	      printf "@{<C.G_Blue_B>I assume a literal@}";
-	      printf "(%s) %a@]@." s Literal.LT.print a;
+	      printf "(%s : %s) %a@]@." n s Literal.LT.print a;
 	      printf "================================================@.@."
 		
 	  | F.Skolem{F.sko_subst=(s,s_ty); sko_f=f} ->
@@ -96,32 +101,34 @@ module Print = struct
       end
 
   let unsat () = 
-    if debug_sat then printf "@[@{<C.Bold>[sat]@} @{<C.G_Red_B>unsat@}@]@."
+    if debug_sat () then printf "@[@{<C.Bold>[sat]@} @{<C.G_Red_B>unsat@}@]@."
 
   let mround s = 
-    if debug_sat then 
+    if debug_sat () then 
       printf "@[@{<C.Bold>[sat]@} matching round of size %d@]@." s
 
   let decide f = 
-    if debug_sat then 
+    if debug_sat () then 
       printf "@[@{<C.Bold>[sat]@} @{<C.G_Green>I decide@} on %a@]@." F.print f
 
   let backtracking f = 
-    if debug_sat then 
+    if debug_sat () then 
       printf "@[@{<C.Bold>[sat]@} @{<C.G_Green>I backtrack@} on %a@]@." 
 	F.print f
 
   let backjumping f = 
-    if debug_sat then 
+    if debug_sat () then 
       (printf "@[@{<C.Bold>[sat]@} @{<C.G_Green>I don't consider the case @}";
        printf "%a@]@." F.print f)
        
-  let elim _ _ = if debug_sat && verbose then printf "@[@{<C.Bold>[elim]@}@."
+  let elim _ _ = 
+    if debug_sat () && verbose () then printf "@[@{<C.Bold>[elim]@}@."
 
-  let red _ _ = if debug_sat && verbose then printf "@[@{<C.Bold>[red]@}@."
+  let red _ _ = 
+    if debug_sat () && verbose () then printf "@[@{<C.Bold>[red]@}@."
 
   let delta d = 
-    if debug_sat && verbose then begin
+    if debug_sat () && verbose () then begin
       printf "@[@{<C.Bold>[sat]@} - Delta ---------------------]@.";
       List.iter (fun (f1, f2, ex) ->
 	printf "(%a or %a), %a@." F.print f1.f F.print f2.f Ex.print ex) d;
@@ -129,27 +136,34 @@ module Print = struct
     end
     
   let gamma g =
-    if debug_sat then begin
+    if debug_sat () then begin
       printf "@[@{<C.Bold>[sat]@} - GAMMA ---------------------]@.";
       MF.iter (fun f ex ->
 	printf "%a \t->\t%a@." F.print f Ex.print ex) g;
       printf "@[@{<C.Bold>[sat]@} --------------------- GAMMA -]@."
-      
     end
-      
+
+  let bottom classes =
+    if bottom_classes () then
+      printf "bottom:%a\n@." Term.print_taged_classes classes      
 
 end
 
 (* matching part of the solver *)
 
-let add_terms env s goal age lem = 
+let add_terms env s goal age lem terms =
+  !Options.timer_start Timers.TMatch;
   let infos = { 
     Matching.term_age = age ; 
     term_from_goal = goal ;
-    term_orig = lem ;
+    term_from_formula = lem ;
+    term_from_terms = terms
   }
   in
-  { env with matching = Term.Set.fold (MM.add_term infos) s env.matching }
+  let env = 
+    { env with matching = Term.Set.fold (MM.add_term infos) s env.matching } in
+  !Options.timer_pause Timers.TMatch;
+  env
 
 exception EnoughLemmasAlready of t * int
 
@@ -196,74 +210,126 @@ let mtriggers env formulas max_size =
       formulas (env, max_size)
   with EnoughLemmasAlready(env, max_size) -> env, max_size
 
-let new_facts mode env = 
+let add_instance_info env orig = 
+  match F.view orig with
+    | F.Lemma _ -> env.add_inst orig
+    | _ -> ()
+
+let new_facts goal_directed env = 
   List.fold_left
     (fun acc ({Matching.trigger_formula=f; trigger_query = guard; 
-	       trigger_age=age; trigger_dep=dep }, subst_list) ->
+	       trigger_age=age; trigger_dep=dep; trigger_orig=orig }, 
+	      subst_list) ->
        List.fold_left
-	 (fun acc {Matching.sbt=s;gen=g;goal=b} ->
+	 (fun acc 
+	    {Matching.sbt = s; 
+	     gen = g; 
+	     goal = b; 
+	     s_term_orig = torig;
+	     s_lem_orig = lorig; } ->
 	    match guard with
 	      | Some a when 
 		  CcX.query (Literal.LT.apply_subst s a) env.tbox = No -> acc
 	      | _ ->
-		  if mode && not b then acc
+		  if goal_directed && not b then acc
 		  else
-		    begin
+		    try
 		      let nf = F.apply_subst s f in
 		      if MF.mem nf env.gamma then acc else
 			let p = 
-			  {f=nf;age=1+(max g age);name=Some f;mf=true;gf=b} in
+			  { f = nf;
+			    age = 1+(max g age);
+			    mf = true;
+			    gf = b;
+			    name = Some lorig;
+			    from_terms = torig;
+			    inv = false
+			  } in
+			add_instance_info env orig;
 			(p,dep)::acc
-		    end
+		    with Exit -> acc
 	 ) 
 	 acc subst_list
     )
     [] (MM.query env.matching env.tbox)
 
 
-let mround predicate mode env max_size =
-  if rules = 2 then fprintf fmt "[rule] TR-Sat-Mround@.";
-  let round mode =
-    Print.mround max_size;
-    let axioms = if predicate then env.definitions else env.lemmas in
-    let env, max_size = mtriggers env axioms max_size in
-    let rec bouclage n (env, lf) = 
-      if n <=0 then (env, lf)
-      else 
-        let env = 
-	  List.fold_left 
-	    (fun env (f,_) -> add_terms env (F.terms f.f) mode f.age None)
-	    env lf
-        in
-        bouclage (n-1) (env, (new_facts mode env))
-    in
-    let _, lf = bouclage Options.bouclage (env, []) in
-    max_size, lf 
+let sort_facts =
+  let rec size f = match F.view f with
+    | F.Unit(f1,f2) -> max (size f1) (size f2)
+    | _             -> F.size f
   in
-  let max_size, lf = round (mode || Options.goal_directed) in 
-  if Options.goal_directed && lf = [] then round false 
-  else max_size, lf
-  
+  fun lf -> List.fast_sort (fun (p1,_) (p2,_) -> size p1.f - size p2.f) lf
 
-let extract_model t = 
+type select = Select_predicates | Select_lemmas | Select_inversions 
+ 
+(* predicate = true : consider only predicates for matching
+   goal_directed = true : match only with terms from goal *)
+
+let mround select ~goal_directed env max_size =
+  if rules () = 2 then fprintf fmt "[rule] TR-Sat-Mround@.";
+  !Options.timer_start Timers.TMatch;
+  Print.mround max_size;
+  let axs = match select with
+    | Select_predicates -> env.definitions
+    | Select_lemmas -> env.lemmas
+    | Select_inversions -> env.inversions
+  in
+  let env, max_size = mtriggers env axs max_size in
+  let lf = new_facts goal_directed env in
+  !Options.timer_pause Timers.TMatch;
+  max_size, sort_facts lf
+
+let is_literal f = match F.view f with F.Literal _ -> true | _ -> false
+
+let extract_prop_model t = 
   let s = ref SF.empty in
   MF.iter 
     (fun f _ -> 
-       let lbl = F.label f in
-       if not (Hstring.equal Hstring.empty lbl) then
+       if (complete_model () && is_literal f) || F.is_in_model f then
 	 s := SF.add f !s
     ) 
     t.gamma;
   !s
 
-let print_model fmt s = 
-  SF.iter (fprintf fmt "%a\n" F.print) s
-  
+let print_prop_model fmt s =
+  SF.iter (fprintf fmt "\n %a" F.print) s
+
+let print_model ~header fmt t =
+  Format.print_flush ();
+  if header then fprintf fmt "\nModel\n@.";
+  let pm = extract_prop_model t in
+  if not (SF.is_empty pm) then begin
+    fprintf fmt "Propositional:";
+    print_prop_model fmt pm;
+    fprintf fmt "\n@.";
+  end;
+  CcX.print_model fmt t.tbox
+
+
+let _ =
+  if not (model ()) then
+    try
+      Sys.set_signal Sys.sigalrm
+	(Sys.Signal_handle (fun _ -> !timeout ()))
+    with Invalid_argument _ -> ()
+
+let refresh_model_handler =
+  if model () then
+    fun t ->
+      try
+	Sys.set_signal Sys.sigalrm
+	  (Sys.Signal_handle (fun _ ->
+	    printf "%a@." (print_model ~header:true) t;
+	    !timeout ()))
+      with Invalid_argument _ -> ()
+  else fun _ -> ()
+
 (* sat-solver *)
 
 let elim {f=f} env = 
   if MF.mem f env.gamma then
-    (if rules = 2 then fprintf fmt "[rule] TR-Sat-Bcp-Elim-1@.";
+    (if rules () = 2 then fprintf fmt "[rule] TR-Sat-Bcp-Elim-1@.";
      true)
   else
     let el = match F.view f with 
@@ -271,7 +337,7 @@ let elim {f=f} env =
       | _ -> false
     in
     if el then 
-      if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-Let@.";
+      if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-Let@.";
     el
 
 let size_formula = 1_000_000
@@ -279,8 +345,8 @@ let size_formula = 1_000_000
 let red {f=f} env = 
   let nf = F.mk_not f in
   try 
-    let r = Yes (MF.find nf env.gamma) in
-    if rules = 2 then fprintf fmt "[rule] TR-Sat-Bcp-Red-1@.";
+    let r = Yes (MF.find nf env.gamma, CcX.cl_extract env.tbox) in
+    if rules () = 2 then fprintf fmt "[rule] TR-Sat-Bcp-Red-1@.";
     r
   with Not_found -> 
     let r = match F.view nf with
@@ -288,13 +354,13 @@ let red {f=f} env =
       | _ -> No
     in
     (match r with 
-      |	Yes _ -> if rules = 2 then fprintf fmt "[rule] TR-Sat-Bcp-Red-2@."
+      |	Yes _ -> if rules () = 2 then fprintf fmt "[rule] TR-Sat-Bcp-Red-2@."
       | No -> ());
     r
 	
 
 let pred_def env f = 
-  let ff = {f=f;age=0;name=None;mf=false;gf=false} in
+  let ff = {f=f;age=0;name=None;mf=false;gf=false; from_terms=[]; inv=false} in
   Print.assume ff Explanation.empty;
   { env with definitions = MF.add f (0,Ex.empty) env.definitions }
 
@@ -302,11 +368,11 @@ let pred_def env f =
 
 let add_dep f dep =
   match F.view f with 
-    | F.Literal _ when proof -> 
+    | F.Literal _ when proof () -> 
       if not (Ex.mem_as_bj f dep) then
 	Ex.union (Ex.singleton ~bj:false f) dep
       else dep
-    | F.Clause _ when proof -> 
+    | F.Clause _ when proof () -> 
 	Ex.union (Ex.singleton ~bj:false f) dep
     | _ -> dep
   
@@ -314,12 +380,14 @@ let add_dep f dep =
 let rec add_dep_of_formula f dep =
   let dep = add_dep f dep in
   match F.view f with 
-    | F.Unit (f1, f2) when proof ->
+    | F.Unit (f1, f2) when proof () ->
       add_dep_of_formula f2 (add_dep_of_formula f1 dep)
     | _ -> dep
 
 
-let rec assume env ({f=f;age=age;name=lem;mf=mf;gf=gf} as ff ,dep) =
+let rec assume env ({f=f;age=age;name=lem;mf=mf;gf=gf;inv=inv} as ff ,dep) =
+  refresh_model_handler env;
+  !Options.thread_yield ();
   try
     let dep = add_dep f dep in
     let dep_gamma = add_dep_of_formula f dep in
@@ -327,12 +395,12 @@ let rec assume env ({f=f;age=age;name=lem;mf=mf;gf=gf} as ff ,dep) =
 	 (* fprintf fmt "ass:%a %a @." 
 	    F.print (F.mk_not f) Ex.print dep_gamma; *)
        let ex_nf = MF.find (F.mk_not f) env.gamma in
-       if rules = 2 then fprintf fmt "[rule] TR-Sat-Conflict-1@.";
-       raise (IUnsat (Ex.union dep_gamma ex_nf))
+       if rules () = 2 then fprintf fmt "[rule] TR-Sat-Conflict-1@.";
+       raise (IUnsat (Ex.union dep_gamma ex_nf, CcX.cl_extract env.tbox))
      with Not_found -> ());
     if MF.mem f env.gamma then
       begin
-	if rules = 2 then fprintf fmt "[rule] TR-Sat-Remove@.";
+	if rules () = 2 then fprintf fmt "[rule] TR-Sat-Remove@.";
 	env
       end
     else 
@@ -341,45 +409,49 @@ let rec assume env ({f=f;age=age;name=lem;mf=mf;gf=gf} as ff ,dep) =
 	if size > size_formula then env
 	else
 	  let env =
-	    if mf && glouton  && size < size_formula then 
-	      add_terms env (F.terms f) gf age lem else env in
+	    if mf && glouton () && size < size_formula then 
+	      add_terms env (F.terms f) gf age lem ff.from_terms else env in
 	  let env = { env with gamma = MF.add f dep_gamma env.gamma } in
 	  Print.assume ff dep;
 	  match F.view f with
 	    | F.Unit (f1, f2) ->
-	      if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-U@.";
-	      let env = assume env 
-		({ f = f1; age = age; name = lem; mf = mf; gf = gf }, dep) in
-	      assume env 
-		({ f = f2; age = age; name = lem; mf = mf; gf = gf }, dep) 
+	      if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-U@.";
+	      let env = assume env ({ ff with f = f1}, dep) in
+	      assume env ({ ff with f = f2}, dep) 
 	    | F.Clause(f1,f2) -> 
-	      if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-C@.";
-		let p1 = {f=f1;age=age;name=lem;mf=mf;gf=gf} in
-		let p2 = {f=f2;age=age;name=lem;mf=mf;gf=gf} in
+	      if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-C@.";
+		let p1 = { ff with f=f1 } in
+		let p2 = { ff with f=f2 } in
 		bcp { env with delta = (p1,p2,dep)::env.delta }
 
-	    | F.Lemma _ ->
-	      if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-Ax@.";
+	    | F.Lemma l ->
+	      if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-Ax@.";
 		let age , dep = 
 		  try 
 		    let age' , dep' = MF.find f env.lemmas in
 		    min age age' , Ex.union dep dep' 
 		  with Not_found -> age , dep 
 		in
-		bcp { env with lemmas=MF.add f (age,dep) env.lemmas }
+		let env = 
+		  if inv then 
+		    { env with inversions=MF.add f (age,dep) env.inversions }
+		  else
+		    { env with lemmas=MF.add f (age,dep) env.lemmas }
+		in
+		bcp env
 
 	    | F.Literal a ->
-		if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-Lit@.";
+		if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-Lit@.";
 		let env = 
 		  if mf && size < size_formula then 
-		    add_terms env (A.LT.terms_of a) gf age lem
+		    add_terms env (A.LT.terms_of a) gf age lem ff.from_terms
 		  else env 
 		in
 		let tbox, new_terms, cpt = CcX.assume a dep env.tbox in
-		let env = add_terms env new_terms gf age lem in
+		let env = add_terms env new_terms gf age lem ff.from_terms in
 		steps := Int64.add (Int64.of_int cpt) !steps;
-		if stepsb <> -1 
-		  && Int64.compare !steps (Int64.of_int stepsb) > 0 then 
+		if stepsb () <> -1 
+		  && Int64.compare !steps (Int64.of_int (stepsb ())) > 0 then 
 		  begin 
 		    printf "Steps limit reached: %Ld@." !steps;
 		    exit 1
@@ -388,25 +460,24 @@ let rec assume env ({f=f;age=age;name=lem;mf=mf;gf=gf} as ff ,dep) =
 		bcp env
 
 	    | F.Skolem{F.sko_subst=sigma; sko_f=f} -> 
-	      if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-Sko@.";
+	      if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-Sko@.";
 		let f' = F.apply_subst sigma f in
-		assume env ({f=f';age=age;name=lem;mf=mf;gf=gf},dep)
+		assume env ({ ff with f=f'},dep)
 
             | F.Let {F.let_var=lvar; let_term=lterm; let_subst=s; let_f=lf} ->
-	      if rules = 2 then fprintf fmt "[rule] TR-Sat-Assume-Let@.";
+	      if rules () = 2 then fprintf fmt "[rule] TR-Sat-Assume-Let@.";
                 let f' = F.apply_subst s lf in
 		let id = F.id f' in
                 let v = Symbols.Map.find lvar (fst s) in
                 let env = assume env 
-		  ({f=F.mk_lit (A.LT.make (A.Eq(v,lterm))) id;
-		    age=age;name=lem;mf=mf;gf=gf},dep) 
+		  ({ ff with f=F.mk_lit (A.LT.make (A.Eq(v,lterm))) id},dep) 
 		in
-                assume env ({f=f';age=age;name=lem;mf=mf;gf=gf},dep)
+                assume env ({ ff with f=f' },dep)
       end
-  with Exception.Inconsistent expl -> 
-    if debug_sat then fprintf fmt "inconsistent %a@." Ex.print expl;
-    if rules = 2 then fprintf fmt "[rule] TR-Sat-Conflict-2@."; 
-    raise (IUnsat expl)
+  with Exception.Inconsistent (expl, classes) -> 
+    if debug_sat () then fprintf fmt "inconsistent %a@." Ex.print expl;
+    if rules () = 2 then fprintf fmt "[rule] TR-Sat-Conflict-2@.";
+    raise (IUnsat (expl, classes))
     
 and bcp env =
   let cl , u = 
@@ -417,16 +488,16 @@ and bcp env =
 	 else 
            (Print.red f1 f2;
 	   match red f1 env with
-	     | Yes d1 -> begin
+	     | Yes (d1, c1) -> begin
 		 match red f2 env with
-		   | Yes d2 -> 
+		   | Yes (d2, c2) -> 
 		       let expl = Ex.union (Ex.union d d1) d2 in
-		       raise (Exception.Inconsistent expl)
+		       raise (Exception.Inconsistent (expl, c1@c2))
 		   | No -> (cl,(f2,Ex.union d d1)::u)
 	       end
 	     | No -> 
 		 match red f2 env with
-		     Yes d2 -> (cl,(f1,Ex.union d d2)::u)
+		     Yes (d2, _) -> (cl,(f1,Ex.union d d2)::u)
 		   | No -> fd::cl , u)
       ) ([],[]) env.delta
   in
@@ -434,82 +505,132 @@ and bcp env =
     
 let rec unsat_rec env fg stop max_size = 
   try
-    if stop < 0 then raise I_dont_know;
+    if stop < 0 then raise (I_dont_know env);
     back_tracking (assume env fg) stop max_size
-  with IUnsat d-> 
-    Print.unsat (); d
+  with IUnsat (d, classes) ->
+    Print.bottom classes;
+    Print.unsat (); 
+    d
 
-and back_tracking env stop max_size =  match env.delta with
-    []  when stop >= 0  -> 
-      let _ , l2 = mround true false env max_max_size in 
+and back_tracking env stop max_size = match env.delta with
+  | []  when stop >= 0  -> 
+      let _ , l2 = 
+	mround Select_predicates ~goal_directed:false env max_max_size in 
       let env = List.fold_left assume env l2 in
 
-      let max_size , l1 = mround false false env max_size in 
+      let max_size , l1 = 
+	mround Select_lemmas ~goal_directed:false env max_size in 
+
       let env = List.fold_left assume env l1 in
 
       let env = 
 	List.fold_left 
-	  (fun env ({f=f; age=g; name=lem; gf=gf},_) -> 
-	     add_terms env (F.terms f) gf g lem) env l1 
+	  (fun env ({f=f; age=g; name=lem; gf=gf} as ff,_) -> 
+	     add_terms env (F.terms f) gf g lem ff.from_terms) env l1 
       in
       (match l1, l2 with
-	 | [], [] -> 
-	     let m = extract_model env in
-	     if all_models then 
+	 | [], [] ->
+	     if all_models () then 
 	       begin
+		 let m = extract_prop_model env in
 		 Format.printf "--- SAT ---\n";
-		 Format.printf "%a@." print_model m;
-		 raise (IUnsat (Ex.make_deps m))
+		 Format.printf "%a@." print_prop_model m;
+		 raise (IUnsat (Ex.make_deps m, []))
 	       end;
 	     raise (Sat env)
 	 | l1, l2 -> 
 	     back_tracking 
 	       (List.fold_left assume  (List.fold_left assume env l2) l1) 
 	       (stop-1) (max_size + b_max_size))
-  | [] -> 
-      raise I_dont_know
+  | [] ->
+      raise (I_dont_know env)
   | (a,b,d)::l ->
       let {f=f;age=g;name=lem;mf=mf} = a in
       Print.decide f;
       let dep = unsat_rec {env with delta=l} (a,Ex.singleton f) stop max_size in
-      if debug_sat then fprintf fmt "unsat_rec : %a@." Ex.print dep;
+      if debug_sat () then fprintf fmt "unsat_rec : %a@." Ex.print dep;
       try
 	let dep' = Ex.remove f dep in
 	Print.backtracking (F.mk_not f);
-	if rules = 2 then fprintf fmt "[rule] TR-Sat-Decide@.";
+	if rules () = 2 then fprintf fmt "[rule] TR-Sat-Decide@.";
 	unsat_rec
 	  (assume {env with delta=l} (b, Ex.union d dep'))
 	  ({a with f=F.mk_not f},dep') stop max_size
       with Not_found -> 
 	Print.backjumping (F.mk_not f);
-	if rules = 2 then fprintf fmt "[rule] TR-Sat-Backjumping@.";
-	dep 
+	if rules () = 2 then fprintf fmt "[rule] TR-Sat-Backjumping@.";
+	dep
 	
-let unsat env fg = 
+let unsat env fg =
   try
-    let env = assume env (fg,Ex.empty) in
-    let env = add_terms env (F.terms fg.f) fg.gf fg.age fg.name in
+    let env = assume env (fg, Ex.empty) in
+    let env = 
+      if not fg.mf then env
+      else add_terms env (F.terms fg.f) fg.gf fg.age fg.name fg.from_terms
+    in
 
-    let _ , l = mround true false env max_max_size in
+    let _ , l = 
+      mround Select_inversions ~goal_directed:false env max_max_size in
     let env = List.fold_left assume env l in
 
-    let _ , l = mround false true env max_max_size in
+    let _ , l = 
+      mround Select_predicates ~goal_directed:false env max_max_size in
     let env = List.fold_left assume env l in
 
-    back_tracking env stopb 100
-  with IUnsat dep -> Print.unsat ();dep
+    let _ , l = 
+      mround Select_lemmas ~goal_directed:true env max_max_size in
+    let env = List.fold_left assume env l in
+
+    back_tracking env (stopb ()) 100
+  with IUnsat (dep, classes) ->
+    Print.bottom classes;
+    Print.unsat ();
+    dep
 
 let assume env fg = 
-  try assume env (fg,Ex.empty) with IUnsat d -> raise (Unsat d)
+  try assume env (fg,Ex.empty) 
+  with IUnsat (d, classes) ->
+    Print.bottom classes;
+    raise (Unsat d)
 
-let empty = { 
+let unsat env fg = 
+  if !profiling then
+    try 
+      !Options.timer_start Timers.TSat;
+      let env = unsat env fg in
+      !Options.timer_pause Timers.TSat;
+      env
+    with e -> 
+      !Options.timer_pause Timers.TSat;
+      raise e
+  else unsat env fg
+
+let assume env fg = 
+  if !profiling then
+    try 
+      !Options.timer_start Timers.TSat;
+      let env = assume env fg in
+      !Options.timer_pause Timers.TSat;
+      env
+    with e -> 
+      !Options.timer_pause Timers.TSat;
+      raise e
+  else assume env fg
+
+
+let empty () = { 
   gamma = MF.empty;
   delta = [] ;
   tbox = CcX.empty (); 
   lemmas = MF.empty ; 
+  inversions = MF.empty ; 
   matching = MM.empty;
-  definitions = MF.empty
+  definitions = MF.empty;
+  add_inst = fun _ -> ();
 } 
 
+let empty_with_inst add_inst =
+  { (empty ()) with add_inst = add_inst }
+
 let start () = steps := 0L
 let stop () = !steps
