@@ -85,7 +85,7 @@ module Make(X : ALIEN) = struct
       | Alien _ , Cons _   -> 1
       | Cons _  , Alien _  -> -1
     
-  let compare x y = compare (embed x) (embed y)
+  let compare x y = compare_mine (embed x) (embed y)
 
   let hash = function
     | Cons (h, ty) -> Hstring.hash h + 19 * Ty.hash ty
@@ -112,23 +112,42 @@ module Make(X : ALIEN) = struct
       | Cons(c1,_) , Cons(c2,_) -> raise Unsolvable
       | Cons _     , Alien r2   -> [r2,a]
       | Alien r1   , Cons _     -> [r1,b]
-      | Alien _    , Alien _    -> assert false
+      | Alien _    , Alien _    -> 
+        if X.compare a b > 0 then [a,b] else [b,a]
     
-  let solve a b = 
-    if debug_sum then fprintf fmt "[Sum] we solve %a = %a@."  
+  let solve_bis a b =
+    if debug_sum () then fprintf fmt "[Sum] we solve %a = %a@."  
       X.print a X.print b;
     try
       let res = solve a b in
-      if debug_sum then (match res with
+      if debug_sum () then (match res with
           [p,v] -> fprintf fmt "\twe get: %a |-> %a@." X.print p X.print v
         | []    -> fprintf fmt "\tthe equation is trivial@."
         | _ -> assert false);
       res
     with Unsolvable -> 
-      if debug_sum then fprintf fmt "\tthe equation is unsolvable@.";
+      if debug_sum () then fprintf fmt "\tthe equation is unsolvable@.";
       raise Unsolvable
 
+
+  let abstract_selectors v acc = is_mine v, acc
+
   let term_extract _ = None
+  
+  let solve r1 r2 pb = 
+    {pb with sbt = List.rev_append (solve_bis r1 r2) pb.sbt}
+
+  let solve r1 r2 pb =
+    if !profiling then
+      try 
+	!Options.timer_start Timers.TSum;
+	let res = solve r1 r2 pb in
+	!Options.timer_pause Timers.TSum;
+	res
+      with e -> 
+	!Options.timer_pause Timers.TSum;
+	raise e
+    else solve r1 r2 pb
 
   module Rel = struct
     type r = X.r
@@ -140,19 +159,19 @@ module Make(X : ALIEN) = struct
     module MX = Map.Make(struct type t = X.r include X end)
     module HSS = Set.Make (struct type t=Hs.t let compare = Hs.compare end)
 
-    type t = (HSS.t * Ex.t) MX.t
+    type t = { mx : (HSS.t * Ex.t) MX.t; classes : Term.Set.t list }
 
-    let empty () = MX.empty
+    let empty classes = { mx = MX.empty; classes = classes }
 
     module Debug = struct
 
       let assume bol r1 r2 =
-        if debug_sum then
+        if debug_sum () then
           fprintf fmt "[Sum.Rel] we assume %a %s %a@." 
             X.print r1 (if bol then "=" else "<>") X.print r2
 
         let print_env env =
-          if debug_sum then begin
+          if debug_sum () then begin
             fprintf fmt "--SUM env ---------------------------------@.";
             MX.iter
               (fun r (hss, ex) ->
@@ -166,7 +185,7 @@ module Make(X : ALIEN) = struct
                  end;
                  fprintf fmt " : %a@." Ex.print ex;
                        
-              ) env;
+              ) env.mx;
             fprintf fmt "-------------------------------------------@.";
           end
     end
@@ -180,12 +199,12 @@ module Make(X : ALIEN) = struct
       match sm1, sm2 with
         | Alien r , Cons(h,ty) 
         | Cons (h,ty), Alien r  ->
-          let enum, ex = try MX.find r env with Not_found -> hss, Ex.empty in
+          let enum, ex = try MX.find r env.mx with Not_found -> hss, Ex.empty in
           let enum = HSS.remove h enum in
           let ex = Ex.union ex dep in
-          if HSS.is_empty enum then raise (Inconsistent ex)
+          if HSS.is_empty enum then raise (Inconsistent (ex, env.classes))
           else 
-            let env = MX.add r (enum, ex) env in
+            let env = { env with mx = MX.add r (enum, ex) env.mx } in
             if HSS.cardinal enum = 1 then
               let h' = HSS.choose enum in
               env, (LSem (A.Eq(r, is_mine (Cons(h',ty)))), ex)::eqs
@@ -198,20 +217,20 @@ module Make(X : ALIEN) = struct
       match sm1, sm2 with
         | Alien r, Cons(h,ty) | Cons (h,ty), Alien r  ->
 
-          let enum, ex = try MX.find r env with Not_found -> hss, Ex.empty in
+          let enum, ex = try MX.find r env.mx with Not_found -> hss, Ex.empty in
           let ex = Ex.union ex dep in
-          if not (HSS.mem h enum) then raise (Inconsistent ex);
-	  MX.add r (HSS.singleton h, ex) env, eqs
+          if not (HSS.mem h enum) then raise (Inconsistent (ex, env.classes));
+	  {env with mx = MX.add r (HSS.singleton h, ex) env.mx} , eqs
             
         | Alien r1, Alien r2   -> 
 
-          let enum1,ex1 = try MX.find r1 env with Not_found -> hss, Ex.empty in
-          let enum2,ex2 = try MX.find r2 env with Not_found -> hss, Ex.empty in
+          let enum1,ex1 = try MX.find r1 env.mx with Not_found -> hss, Ex.empty in
+          let enum2,ex2 = try MX.find r2 env.mx with Not_found -> hss, Ex.empty in
           let ex = Ex.union dep (Ex.union ex1 ex2) in
           let diff = HSS.inter enum1 enum2 in 
-          if HSS.is_empty diff then raise (Inconsistent ex);
-          let env = MX.add r1 (diff, ex) env in
-          let env = MX.add r2 (diff, ex) env in
+          if HSS.is_empty diff then raise (Inconsistent (ex, env.classes));
+          let mx = MX.add r1 (diff, ex) env.mx in
+          let env = {env with mx = MX.add r2 (diff, ex) mx } in
           if HSS.cardinal diff = 1 then
             let h' = HSS.choose diff in
             let ty = X.type_info r1 in
@@ -220,7 +239,8 @@ module Make(X : ALIEN) = struct
 
         |  _ -> env, eqs
 
-    let assume env la ~are_eq ~are_neq ~class_of = 
+    let assume env la ~are_eq ~are_neq ~class_of ~classes =
+      let env = { env with classes = classes } in
       let aux bol r1 r2 dep env eqs = function
         | None     -> env, eqs
         | Some hss -> 
@@ -257,31 +277,63 @@ module Make(X : ALIEN) = struct
            else match acc with
 	     | Some (n,r,hs) when n <= sz -> acc
 	     | _ -> Some (sz, r, HSS.choose hss)
-        ) env None 
+        ) env.mx None 
       in
       match acc with 
         | Some (n,r,hs) -> 
 	    let r' = is_mine (Cons(hs,X.type_info r)) in
-	    if debug_sum then
+	    if debug_sum () then
 	      fprintf fmt "[case-split] %a = %a@." X.print r X.print r';
 	    [A.Eq(r, r'), Ex.empty, Num.Int n]
         | None -> 
-	    if debug_sum then fprintf fmt "[case-split] sum: nothing@.";
+	    if debug_sum () then fprintf fmt "[case-split] sum: nothing@.";
 	    []
       
 
-    let query env a_ex ~are_eq ~are_neq ~class_of =
-      try ignore(assume env [a_ex] ~are_eq ~are_neq ~class_of); Sig.No
-      with Inconsistent expl -> Sig.Yes expl          
+    let query env a_ex ~are_eq ~are_neq ~class_of ~classes =
+      try ignore(assume env [a_ex] ~are_eq ~are_neq ~class_of ~classes); Sig.No
+      with Inconsistent (expl, classes) -> Sig.Yes (expl, classes)
+
+    let add env r =
+      match embed r, values_of r with
+	| Alien r, Some hss -> 
+          if MX.mem r env.mx then env else 
+            { env with mx = MX.add r (hss, Ex.empty) env.mx }
+	| _ -> env
+
+
+  let assume env la ~are_eq ~are_neq ~class_of ~classes =
+    if !profiling then
+      try 
+	!Options.timer_start Timers.TSum;
+	let res =assume env la ~are_eq ~are_neq ~class_of ~classes in
+	!Options.timer_pause Timers.TSum;
+	res
+      with e -> 
+	!Options.timer_pause Timers.TSum;
+	raise e
+    else assume env la ~are_eq ~are_neq ~class_of ~classes
+
+  let query env la ~are_eq ~are_neq ~class_of ~classes =
+    if !profiling then
+      try 
+	!Options.timer_start Timers.TSum;
+	let res = query env la ~are_eq ~are_neq ~class_of ~classes in
+	!Options.timer_pause Timers.TSum;
+	res
+      with e -> 
+	!Options.timer_pause Timers.TSum;
+	raise e
+    else query env la ~are_eq ~are_neq ~class_of ~classes
+
 
-    let add env r = match embed r, values_of r with
-      | Alien r, Some hss -> 
-          if MX.mem r env then env else 
-            MX.add r (hss, Ex.empty) env
 
-      | _ -> env
 
     let instantiate env _ _ _ _ = env, []
 
+    let print_model _ _ _ = ()
+
+    let new_terms env = Term.Set.empty
+
   end
 end
