@@ -61,27 +61,44 @@ let rec print fmt t =
     | Sy.Op Sy.Extract, [e1; e2; e3] ->
 	fprintf fmt "%a^{%a,%a}" print e1 print e2 print e3
 
+    | Sy.Op (Sy.Access field), [e] ->
+	fprintf fmt "%a.%s" print e (Hstring.view field)
+
+    | Sy.Op (Sy.Record), _ ->
+        begin match ty with
+	  | Ty.Trecord {Ty.lbs=lbs} ->
+	    assert (List.length l = List.length lbs);
+	      fprintf fmt "{";
+	      ignore (List.fold_left2 (fun first (field,_) e -> 
+		fprintf fmt "%s%s = %a"  (if first then "" else "; ")
+		  (Hstring.view field) print e;
+		false
+	      ) true lbs l);
+	      fprintf fmt "}";
+	  | _ -> assert false
+	end
+
     | Sy.Op op, [e1; e2] -> 
 	fprintf fmt "(%a %a %a)" print e1 Sy.print x print e2
 
     | _, [] -> 
-        if Options.debug then
+        if Options.debug () then
           fprintf fmt "%a:%a" Sy.print x Ty.print ty
         else
           fprintf fmt "%a" Sy.print x
           
     | _, _ ->
-        if Options.debug then 
+        if Options.debug () then 
           fprintf fmt "%a(%a):%a" Sy.print x print_list l Ty.print ty
         else 
           fprintf fmt "%a(%a)" Sy.print x print_list l
 
-and print_list fmt = function
+and print_list_sep sep fmt = function
   | [] -> ()
   | [t] -> print fmt t
-  | t::l -> Format.fprintf fmt "%a,%a" print t print_list l
-
+  | t::l -> Format.fprintf fmt "%a%s%a" print t sep (print_list_sep sep) l
 
+and print_list fmt = print_list_sep "," fmt
 
 
 (* fresh variables must be smaller than problem's variables.
@@ -104,6 +121,11 @@ let make s l ty = T.hashcons {f=s;xs=l;ty=ty;tag=0 (* dumb_value *) }
 
 let fresh_name ty = make (Sy.name (Common.fresh_string())) [] ty
 
+let is_fresh t = 
+  match view t with
+    | {f=Sy.Name(hs,_);xs=[]} -> Common.is_fresh_string (Hstring.view hs)
+    | _ -> false
+
 let shorten t = 
   let {f=f;xs=xs;ty=ty} = view t in
   make f xs (Ty.shorten ty)
@@ -158,9 +180,28 @@ let vty_of t =
   in
   vty_of Ty.Svty.empty t
 
+module Hsko = Hashtbl.Make(H)
+let gen_sko ty = make (Sy.fresh "@sko") [] ty
+
+let is_skolem_cst v = 
+  try
+    String.sub (Sy.to_string v.f) 0 4 = "_sko"
+  with Invalid_argument _ -> false
+
+let find_skolem = 
+  let hsko = Hsko.create 17 in
+  fun v ty ->
+    if is_skolem_cst v then
+      try Hsko.find hsko v
+      with Not_found -> 
+	let c = gen_sko ty in Hsko.add hsko v c; c
+    else v
+
 let rec apply_subst ((s_t,s_ty) as s) t = 
   let {f=f;xs=xs;ty=ty} = view t in
-  try Sy.Map.find f s_t
+  try 
+    let v = Sy.Map.find f s_t in
+    find_skolem v ty
   with Not_found -> 
     make f (List.map (apply_subst s) xs) (Ty.apply_subst s_ty ty)
 
@@ -181,3 +222,38 @@ let union_subst (s_t1, s_ty1) ((s_t2, s_ty2) as subst) =
 
 let rec subterms acc t = 
   let {xs=xs} = view t in List.fold_left subterms (Set.add t acc) xs
+
+
+module Labels = Hashtbl.Make(H)
+  
+let labels = Labels.create 100007
+  
+let add_label lbl t = 
+  Labels.replace labels t lbl
+  
+let label t = try Labels.find labels t with Not_found -> Hstring.empty
+
+
+let rec is_in_model_rec depth { f = f; xs = xs } =
+  let lb = Symbols.label f in
+  (Common.label_model lb
+   &&
+     (try
+	let md = Scanf.sscanf (Hstring.view lb) "model:%d" (fun x -> x) in
+	depth <= md
+      with Scanf.Scan_failure _ | End_of_file-> true))
+  || 
+    List.exists (is_in_model_rec (depth +1)) xs
+
+let is_in_model t =
+  Common.label_model (label t) || is_in_model_rec 0 t
+
+
+let is_labeled t = not (Hstring.equal (label t) Hstring.empty)
+
+let print_taged_classes fmt =
+  List.iter (fun cl -> 
+    let cl = List.filter is_labeled (Set.elements cl) in
+    if cl <> [] then
+    fprintf fmt "\n{ %a }" (print_list_sep " , ") cl)
+
