@@ -40,7 +40,7 @@ val real : string -> t
 val bitv : string -> Ty.t -> t
 
 val fresh_name : Ty.t -> t
-
+val is_fresh : t -> bool
 val is_int : t -> bool
 val is_real : t -> bool
 
@@ -60,8 +60,14 @@ val fold_subst_term : (Symbols.t -> t -> 'b -> 'b) -> subst -> 'b -> 'b
 
 val union_subst : subst -> subst -> subst
 
+val add_label : Hstring.t -> t -> unit
+val label : t -> Hstring.t
+val is_in_model : t -> bool
+
 val print : Format.formatter -> t -> unit
 val print_list : Format.formatter -> t list -> unit
+val print_list_sep : string -> Format.formatter -> t list -> unit
+val print_taged_classes : Format.formatter -> Set.t list -> unit
 
 val dummy : t 
 
