@@ -87,6 +87,7 @@ let rec depth_tterm t =
 	max (depth_tterm t1) (max (depth_tterm t2) (depth_tterm t3))
     | TTlet (s, t1, t2) ->
 	max (depth_tterm t1 + 1) (depth_tterm t2)
+    | TTnamed (_, t) -> depth_tterm t
 
 exception Out of int
 
@@ -194,6 +195,9 @@ let rec compare_tterm t1 t2 =
 	if c<>0 then c else
 	  let c = compare_tterm t1 u1 in
 	  if c<>0 then c else compare_tterm u1 u2
+    | TTnamed (_, t), _ -> compare_tterm t t2
+    | _, TTnamed (_, t) -> compare_tterm t1 t
+
 
 let compare_tterm_list tl2 tl1 =
   let l1 = List.map depth_tterm tl1 in
@@ -216,10 +220,138 @@ let compare_tterm_list tl2 tl1 =
 	 if c <> 0 then raise (Out c)
       ) tl1 tl2; 0 
   with Out c -> c end
-    
+
+
+module Uniq_sort = struct
+  let rec merge cmp l1 l2 =
+    match l1, l2 with
+      | [], l2 -> l2
+      | l1, [] -> l1
+      | h1 :: t1, h2 :: t2 ->
+	let c = cmp h1 h2 in
+	if c = 0 then h1 :: merge cmp t1 t2
+	else if c < 0 then h1 :: merge cmp t1 l2
+	else h2 :: merge cmp l1 t2
+
+
+  let rec chop k l =
+    if k = 0 then l else begin
+      match l with
+	| x::t -> chop (k-1) t
+	| _ -> assert false
+    end
+  ;;
+
+  let stable_sort cmp l =
+    let rec rev_merge l1 l2 accu =
+      match l1, l2 with
+	| [], l2 -> List.rev_append l2 accu
+	| l1, [] -> List.rev_append l1 accu
+	| h1::t1, h2::t2 ->
+          let c = cmp h1 h2 in
+	  if c = 0 then rev_merge t1 t2 (h1::accu)
+	  else if c < 0 then rev_merge t1 l2 (h1::accu)
+          else rev_merge l1 t2 (h2::accu)
+    in
+    let rec rev_merge_rev l1 l2 accu =
+      match l1, l2 with
+	| [], l2 -> List.rev_append l2 accu
+	| l1, [] -> List.rev_append l1 accu
+	| h1::t1, h2::t2 ->
+          let c = cmp h1 h2 in
+	  if c = 0 then rev_merge_rev t1 t2 (h1::accu)
+	  else if c > 0 then rev_merge_rev t1 l2 (h1::accu)
+          else rev_merge_rev l1 t2 (h2::accu)
+    in
+    let rec sort n l =
+      match n, l with
+	| 2, x1 :: x2 :: _ ->
+	  let c = cmp x1 x2 in
+	  if c = 0 then [x1] 
+	  else if c < 0 then [x1; x2] else [x2; x1]
+	| 3, x1 :: x2 :: x3 :: _ ->
+	  let c = cmp x1 x2 in
+	  if c = 0 then begin
+	    let c = cmp x2 x3 in
+	    if c = 0 then [x1]
+            else if c <= 0 then [x1; x3]
+            else [x3; x1]
+	  end
+	  else if c < 0 then begin
+	    let c = cmp x2 x3 in
+	    if c = 0 then [x1; x2]
+            else if c < 0 then [x1; x2; x3]
+            else 
+	      let c = cmp x1 x3 in
+	      if c = 0 then [x1; x2]
+	      else if c < 0 then [x1; x3; x2]
+              else [x3; x1; x2]
+	  end else begin
+	    let c = cmp x1 x3 in
+	    if c = 0 then [x2; x1]
+            else if c < 0 then [x2; x1; x3]
+            else
+	      let c = cmp x2 x3 in
+	      if c = 0 then [x2; x1]
+	      else if c < 0 then [x2; x3; x1]
+              else [x3; x2; x1]
+	  end
+	| n, l ->
+	  let n1 = n asr 1 in
+	  let n2 = n - n1 in
+	  let l2 = chop n1 l in
+	  let s1 = rev_sort n1 l in
+	  let s2 = rev_sort n2 l2 in
+	  rev_merge_rev s1 s2 []
+    and rev_sort n l =
+      match n, l with
+	| 2, x1 :: x2 :: _ ->
+	  let c = cmp x1 x2 in
+	  if c = 0 then [x1] 
+	  else if c > 0 then [x1; x2] else [x2; x1]
+	| 3, x1 :: x2 :: x3 :: _ ->
+	  let c = cmp x1 x2 in
+	  if c = 0 then begin
+	    let c = cmp x2 x3 in
+	    if c = 0 then [x1]
+            else if c > 0 then [x1; x3]
+            else [x3; x1]
+	  end
+	  else if c > 0 then begin
+	    let c = cmp x2 x3 in
+	    if c = 0 then [x1; x2]
+            else if c > 0 then [x1; x2; x3]
+            else
+	      let c = cmp x1 x3 in
+	      if c = 0 then [x1; x2]
+	      else if c > 0 then [x1; x3; x2]
+              else [x3; x1; x2]
+	  end else begin
+	    let c = cmp x1 x3 in
+	    if c = 0 then [x2; x1]
+            else if c > 0 then [x2; x1; x3]
+            else
+	      let c = cmp x2 x3 in
+	      if c = 0 then [x2; x1]
+	      else if c > 0 then [x2; x3; x1]
+              else [x3; x2; x1]
+	  end
+	| n, l ->
+	  let n1 = n asr 1 in
+	  let n2 = n - n1 in
+	  let l2 = chop n1 l in
+	  let s1 = sort n1 l in
+	  let s2 = sort n2 l2 in
+	  rev_merge s1 s2 []
+    in
+    let len = List.length l in
+    if len < 2 then l else sort len l
+end
+
+
 
 let at_most n l =
-  let l = List.fast_sort compare_tterm_list l in
+  let l = Uniq_sort.stable_sort compare_tterm_list l in
   let rec atmost acc n l = 
     match n, l with
       | n, _ when n <= 0 -> acc
@@ -231,9 +363,9 @@ let at_most n l =
   List.rev (atmost [] n l)
 
 let is_var t = match t.c.tt_desc with
-  | TTvar _ -> true
-  | TTapp (_,l) -> l=[]
-  | _ -> false
+  | TTvar (Symbols.Var _) -> true
+  | _ -> false (* constant terms such as "logic nil : 'a list" 
+                  are allowed in triggers *)
 
 module SLLT = 
   Set.Make(
@@ -329,9 +461,9 @@ let multi_triggers gopt bv vty trs =
   let lm = List.map (fun (lt, _, _) -> lt) l_parties in
   let mv , mt = List.partition (List.exists is_var) lm in
   let mv , mt = sort mv , sort mt in
-  let lm = if gopt || triggers_var then mt@mv else mt in
-  let m = at_most redondance  lm in
-  at_most redondance m
+  let lm = if gopt || triggers_var () then mt@mv else mt in
+  let m = at_most (redondance ()) lm in
+  at_most (redondance ()) m
 
 let rec vty_ty acc t = 
   let t = Ty.shorten t in
@@ -469,21 +601,31 @@ let filter_good_triggers (bv, vty) =
 
 let make_triggers gopt vterm vtype trs = 
   match List.filter (filter_mono vterm vtype) trs with
-      [] -> 
+    | [] ->
 	multi_triggers gopt vterm vtype trs
     | trs' -> 
-	let f l = at_most redondance (List.map (fun (t, _, _) -> [t]) l) in
+	let f l = at_most (redondance ()) (List.map (fun (t, _, _) -> [t]) l) in
 	let trs_v, trs_nv = List.partition (fun (t, _, _) -> is_var t) trs' in
 	let ll = 
 	  if trs_nv = [] then
-	    if triggers_var || gopt then 
+	    if triggers_var () || gopt then 
 	      f trs_v 
 	    else [] (*multi_triggers vars trs*)
 	  else f trs_nv 
 	in 
-	if glouton then ll@(multi_triggers gopt vterm vtype trs) else ll
+	if glouton () then ll@(multi_triggers gopt vterm vtype trs) else ll
 
-let rec make_rec pol gopt vterm vtype f = 
+let check_triggers trs (bv, vty) =
+  if trs = [] then 
+         failwith "There should be a trigger for every quantified formula in a theory.";
+  List.iter (fun l -> 
+       let s1 = List.fold_left (vars_of_term bv) Vterm.empty l in
+       let s2 = List.fold_left vty_term Vtype.empty l in
+       if not (Vtype.subset vty s2) or not (Vterm.subset bv s1) then 
+         failwith "Triggers of a theory should contain every quantified types
+and variables.") trs; trs
+
+let rec make_rec keep_triggers pol gopt vterm vtype f = 
   let c, trs = match f.c with
     | TFatom {c = (TAfalse | TAtrue)} ->
 	f.c, STRS.empty
@@ -521,13 +663,15 @@ let rec make_rec pol gopt vterm vtype f =
 	  end
     | TFop (OPimp, [f1; f2]) -> 
 	
-	let f1, trs1  = make_rec (neg_pol pol) gopt vterm vtype f1 in
-	let f2, trs2  = make_rec pol gopt vterm vtype f2 in
+	let f1, trs1  = 
+          make_rec keep_triggers (neg_pol pol) gopt vterm vtype f1 in
+	let f2, trs2  = make_rec keep_triggers pol gopt vterm vtype f2 in
 	let trs = STRS.union trs1 trs2 in
 	TFop(OPimp, [f1; f2]), trs
 
     | TFop (OPnot, [f1]) -> 
-	let f1, trs1  = make_rec (neg_pol pol) gopt vterm vtype f1 in
+	let f1, trs1  = 
+          make_rec keep_triggers (neg_pol pol) gopt vterm vtype f1 in
 	TFop(OPnot, [f1]), trs1
 
     (* | OPiff 
@@ -537,7 +681,7 @@ let rec make_rec pol gopt vterm vtype f =
 	let lf, trs = 
 	  List.fold_left
 	    (fun (lf, trs1) f ->
-	       let f, trs2 = make_rec pol gopt vterm vtype f in
+	       let f, trs2 = make_rec keep_triggers pol gopt vterm vtype f in
 	       f::lf, STRS.union trs1 trs2) ([], STRS.empty) lf in
 	TFop(op,List.rev lf), trs
 
@@ -550,10 +694,11 @@ let rec make_rec pol gopt vterm vtype f =
 
 	let vterm'' = Vterm.union vterm vterm' in
 	let vtype'' = Vtype.union vtype vtype' in
-	let f1', trs1 = make_rec pol gopt vterm'' vtype'' f1 in
-	let f2', trs2 = make_rec pol gopt vterm'' vtype'' f2 in
+	let f1', trs1 = make_rec keep_triggers pol gopt vterm'' vtype'' f1 in
+	let f2', trs2 = make_rec keep_triggers pol gopt vterm'' vtype'' f2 in
 	let trs12 = 
-	  if Options.notriggers || qf.qf_triggers = [] then
+	  if keep_triggers then check_triggers qf.qf_triggers (vterm', vtype')
+	  else if Options.notriggers () || qf.qf_triggers = [] then
 	    begin
 	      (make_triggers false vterm' vtype' (STRS.elements trs1))@
 		(make_triggers false vterm' vtype' (STRS.elements trs2))
@@ -588,10 +733,11 @@ let rec make_rec pol gopt vterm vtype f =
 	  List.fold_left 
 	    (fun b (s,_) -> Vterm.add s b) Vterm.empty qf.qf_bvars in
 	let f', trs = 
-	  make_rec pol gopt 
+	  make_rec keep_triggers pol gopt 
 	    (Vterm.union vterm vterm') (Vtype.union vtype vtype') qf.qf_form in
 	let trs' = 
-	  if Options.notriggers || qf.qf_triggers=[] then
+	  if keep_triggers then check_triggers qf.qf_triggers (vterm', vtype')
+	  else if Options.notriggers () || qf.qf_triggers=[] then
 	    make_triggers gopt vterm' vtype' (STRS.elements trs)
 	  else 
 	    let lf = filter_good_triggers (vterm',vtype') qf.qf_triggers in
@@ -607,27 +753,29 @@ let rec make_rec pol gopt vterm vtype f =
 	   | _ -> TFexists r , trs)
 
     | TFlet (up, v, t, f) ->
-	let f, trs = make_rec pol gopt vterm vtype f in 
+	let f, trs = make_rec keep_triggers pol gopt vterm vtype f in 
 	let trs = STRS.union trs (potential_triggers (vterm, vtype) [t]) in
 	(* XXX correct for terms *)
 	TFlet (up, v, t, f), trs
 
     | TFnamed(lbl, f) -> 
-	let f, trs = make_rec pol gopt vterm vtype f in
+	let f, trs = make_rec keep_triggers pol gopt vterm vtype f in
 	TFnamed(lbl, f), trs
   in
   { f with c = c }, trs
 	  
-let make gopt f = match f.c with
+let make keep_triggers gopt f = match f.c with
   | TFforall _ | TFexists _ -> 
-      let f, _ = make_rec Pos gopt Vterm.empty Vtype.empty f in
+      let f, _ = make_rec keep_triggers Pos gopt Vterm.empty Vtype.empty f in
       f
   | _  ->  
       let vty = vty_form Vtype.empty f in
-      let f, trs = make_rec Pos gopt Vterm.empty vty f in
+      let f, trs = make_rec keep_triggers Pos gopt Vterm.empty vty f in
       if Vtype.is_empty vty then f
       else 
 	let trs = STRS.elements trs in
+        if keep_triggers then
+          failwith "No polymorphism in use-defined theories.";
 	let trs = make_triggers gopt Vterm.empty vty trs in
 	{ f with c = TFforall 
 	  {qf_bvars=[]; qf_upvars=[]; qf_triggers=trs; qf_form=f }}
