@@ -44,8 +44,11 @@ module type S = sig
   val are_equal : t -> Term.t -> Term.t -> Sig.answer
   val are_distinct : t -> Term.t -> Term.t -> Sig.answer
   val already_distinct : t -> R.r list -> bool
+  
   val class_of : t -> Term.t -> Term.t list
-
+  val cl_extract : t -> Term.Set.t list
+  val model : t -> 
+    (R.r * Term.t list * (Term.t * R.r) list) list * (Term.t list) list
   val print : Format.formatter -> t -> unit
  
 end
@@ -65,7 +68,7 @@ module Make ( R : Sig.X ) = struct
   module SetF = Formula.Set
     
   module Lit = Literal.Make(struct type t = R.r include R end)
-  module MapL = Lit.Map  
+  module MapL = Lit.Map
 
   module MapR = Map.Make(struct type t = R.r let compare = R.compare end)
     
@@ -133,6 +136,42 @@ module Make ( R : Sig.X ) = struct
     ac_rs = RS.empty
   }
 
+  exception Found_term of T.t
+
+  (* hack: would need an inverse map from semantic values to terms *)
+  let terms_of_distinct env l = match Lit.view l with
+    | Literal.Distinct (false, rl) ->
+        let lt =
+          List.fold_left (fun acc r -> 
+	    try 
+	      let cl = MapR.find r env.classes in
+	      SetT.iter (fun t -> 
+		if R.equal (MapT.find t env.make) r then
+		  raise (Found_term t)) cl;
+	      acc
+	    with 
+	      | Found_term t -> t :: acc
+	      | Not_found -> acc) [] rl
+	in
+	let rec distrib = function
+	  | x :: r -> (distrib r) @ 
+	    (List.map (fun y -> SetT.add x (SetT.singleton y)) r)  
+	  | [] -> []
+	in
+	distrib lt
+
+    | _ -> assert false
+
+
+  let cl_extract env =
+    if bottom_classes () then
+      let classes = MapR.fold (fun _ cl acc -> cl :: acc) env.classes [] in
+      MapR.fold (fun _ ml acc -> 
+	MapL.fold (fun l _ acc -> (terms_of_distinct env l) @ acc) ml acc
+      ) env.neqs classes
+    else []
+	
+
   module Print = struct
 
     let rs_print fmt = SetR.iter (fprintf fmt "\t%a@." R.print)
@@ -176,7 +215,7 @@ module Make ( R : Sig.X ) = struct
       MapR.iter (fun k s -> fprintf fmt "%a -> %a\n" R.print k lm_print s) m
 
     let all fmt env = 
-      if debug_uf then
+      if debug_uf () then
 	begin
 	  fprintf fmt "-------------------------------------------------@.";
 	  fprintf fmt "%a %a %a %a %a" 
@@ -195,15 +234,18 @@ module Make ( R : Sig.X ) = struct
     let mem env t = MapT.mem t env.make
       
     let lookup_by_t t env =
+      !Options.thread_yield ();
       try MapR.find (MapT.find t env.make) env.repr
       with Not_found -> 
-	if debug_uf then fprintf fmt "Uf: Not_found %a@." Term.print t;
+	if debug_uf () then fprintf fmt "Uf: Not_found %a@." Term.print t;
 	assert false (*R.make t, Ex.empty*) (* XXXX *)
 	  
     let lookup_by_r r env = 
+      !Options.thread_yield ();
       try MapR.find r env.repr with Not_found -> r, Ex.empty
 	
     let lookup_for_neqs env r =
+      !Options.thread_yield ();
       try MapR.find r env.neqs with Not_found -> MapL.empty
 	
     let add_to_classes t r classes =  
@@ -216,7 +258,8 @@ module Make ( R : Sig.X ) = struct
       let s2 = try MapR.find nc classes with Not_found -> SetT.empty in
       MapR.remove c (MapR.add nc (SetT.union s1 s2) classes)
 	
-    let add_to_gamma r c gamma = 
+    let add_to_gamma r c gamma =
+      !Options.thread_yield ();
       L.fold_left
 	(fun gamma x -> 
 	  let s = try MapR.find x gamma with Not_found -> SetR.empty in
@@ -232,9 +275,9 @@ module Make ( R : Sig.X ) = struct
 	     try 
 	       let ex2 = MapL.find l1 mapl in
 	       let ex = Ex.union (Ex.union ex1 ex2) dep in (* VERIF *)
-	       if rules = 3 then
+	       if rules () = 3 then
 		 fprintf fmt "[rule] TR-CCX-Congruence-Conflict@.";
-	       raise (Inconsistent ex)
+	       raise (Inconsistent (ex, cl_extract env))
 	     with Not_found -> 
 	       MapL.add l1 (Ex.union ex1 dep) mapl) 
 	  nq_r1 nq_r2
@@ -243,6 +286,7 @@ module Make ( R : Sig.X ) = struct
 
     let disjoint_union l_1 l_2 = 
       let rec di_un (l1,c,l2) (l_1,l_2)= 
+	!Options.thread_yield ();
         match l_1,l_2 with
 	  | [],[] -> l1, c, l2
 	  | l, [] -> di_un (l @ l1,c,l2) ([],[])
@@ -276,7 +320,15 @@ module Make ( R : Sig.X ) = struct
 	      else raise List_minus_exn
       in di_un [] l_1 l_2
         
+    let setrl fmt rls = 
+      SetRL.iter
+        (fun (g,d,_) ->
+          fprintf fmt "%a ----> %a@." Ac.print g R.print d;
+        )rls
+
     let apply_rs r rls = 
+      (*fprintf fmt "apply_rs %a and rs is@. %a@.@." Ac.print r setrl rls;*)
+
       let fp = ref true in
       let r = ref r in
       let ex = ref Ex.empty in
@@ -296,6 +348,7 @@ module Make ( R : Sig.X ) = struct
           with List_minus_exn -> ()
       in
       let rec fixpoint () = 
+	!Options.thread_yield ();
         (try SetRL.iter apply_rule rls with Exit -> ());
 	if !fp then !r, !ex else (fp := true; fixpoint ())
       in fixpoint()
@@ -327,14 +380,14 @@ module Make ( R : Sig.X ) = struct
     let rec canon env r ex_r = 
       let se, sac = filter_leaves r in
       let subst, ex_subst = canon_empty se env in
-      let sac, ex_ac = canon_ac sac env in (* explications? *)
-      let r2 = canon_aux (canon_aux r sac) subst in
+      let subst_ac, ex_ac = canon_ac sac env in (* explications? *)
+      let r2 = canon_aux (canon_aux r subst_ac) subst in
       let ex_r2 = Ex.union (Ex.union ex_r ex_subst) ex_ac in
       if R.equal r r2 then r2, ex_r2 else canon env r2 ex_r2
 
     let normal_form env r =
       let rr, ex = canon env r Ex.empty in
-      if rewriting && verbose then 
+      if rewriting () && verbose () then 
         fprintf fmt "canon %a = %a@." R.print r R.print rr;
       rr,ex
 
@@ -342,24 +395,24 @@ module Make ( R : Sig.X ) = struct
 
 
     let find_or_normal_form env r =
+      !Options.thread_yield ();
       try MapR.find r env.repr with Not_found -> normal_form env r
 
-    let init_leaf env p = 
-      if debug_uf then fprintf fmt "init_leaf: %a@." R.print p;
+    let init_leaf env p =
+      if debug_uf () then fprintf fmt "init_leaf: %a@." R.print p;
       let in_repr = MapR.mem p env.repr in
-      let in_neqs = MapR.mem p env.neqs in
       { env with
 	repr    = 
 	  if in_repr then env.repr 
 	  else MapR.add p (p, Ex.empty) env.repr;
 	classes = 
-	  if in_repr then env.classes
+	  if MapR.mem p env.classes then env.classes
 	  else update_classes p p env.classes;
 	gamma   = 
 	  if in_repr then env.gamma
 	  else add_to_gamma p p env.gamma ;
 	neqs    =  
-	  if in_neqs then env.neqs 
+	  if MapR.mem p env.neqs then env.neqs 
 	  else update_neqs p p Ex.empty env } 
           
     let init_term env t = 
@@ -384,7 +437,7 @@ module Make ( R : Sig.X ) = struct
 	       | l1 , cm , l2 -> 
 		   let rx = R.color {ac with l = Ac.add h (d,1) l1} in
 		   let ry = R.color {g  with l = Ac.add h (v,1) l2} in
-                   if debug_uf then
+                   if debug_uf () then
                      fprintf fmt "[uf] critical pair: %a = %a@." 
                        R.print rx R.print ry;
                    if not (R.equal rx ry) then 
@@ -397,29 +450,40 @@ module Make ( R : Sig.X ) = struct
 	(fun h rls env ->
           SetRL.fold
 	    (fun ((g, d, dep_rl) as rul) env ->
+	      !Options.thread_yield ();
 	      let env = {env with ac_rs = RS.remove_rule rul env.ac_rs} in
 	      let gx = R.color g in
 	      let g2, ex_g2 = normal_form env (Ac.subst p v g) in
 	      let d2, ex_d2 = normal_form env (R.subst p v d) in
-	      if R.equal g2 gx then (* compose *)
-		begin
-		  if R.equal g2 d2 then
-		    Format.eprintf "Compose : %a -> %a sur %a et %a@." 
-		      R.print p R.print v
-		      Ac.print g R.print d;
-                  let ex = Ex.union ex_d2 (Ex.union dep_rl dep) in
-	          {env with ac_rs = RS.add_rule (g,d2, ex) env.ac_rs}
+              if R.compare g2 d2 <= 0 then begin
+                if debug_ac () then
+                  fprintf fmt "[uf] collapse forc par *: %a = %a@."
+		    R.print g2 R.print d2;
+                let ex = Ex.union 
+		  (Ex.union ex_g2 ex_d2) (Ex.union dep_rl dep) in
+                Queue.push (g2, d2, ex) eqs;
+	        env
+              end
+              else
+	        if R.equal g2 gx then (* compose *)
+		  begin
+		    if debug_ac () && R.equal g2 d2 then
+		      Format.eprintf "Compose : %a -> %a on %a and %a@." 
+		        R.print p R.print v
+		        Ac.print g R.print d;
+                    let ex = Ex.union ex_d2 (Ex.union dep_rl dep) in
+	            {env with ac_rs = RS.add_rule (g,d2, ex) env.ac_rs}
 		  end
-	      else (* collapse *)
-                begin
-                  if debug_ac then
-                    fprintf fmt "[uf] collapse: %a = %a@."
-		      R.print g2 R.print d2;
-                  let ex = Ex.union 
-		    (Ex.union ex_g2 ex_d2) (Ex.union dep_rl dep) in
-                  Queue.push (g2, d2, ex) eqs;
-	          env
-                end
+	        else (* collapse *)
+                  begin
+                    if debug_ac () then
+                      fprintf fmt "[uf] collapse: %a = %a@."
+		        R.print g2 R.print d2;
+                    let ex = Ex.union 
+		      (Ex.union ex_g2 ex_d2) (Ex.union dep_rl dep) in
+                    Queue.push (g2, d2, ex) eqs;
+	            env
+                  end
 	    ) rls env
 	) env.ac_rs env
 	
@@ -447,7 +511,8 @@ module Make ( R : Sig.X ) = struct
       let use_p = MapR.find p env.gamma in
       try 
 	let env, tch, neqs_to_up = SetR.fold 
-	  (fun r (env, touched, neqs_to_up) -> 
+	  (fun r (env, touched, neqs_to_up) ->
+	    !Options.thread_yield (); 
 	     let rr, ex = MapR.find r env.repr in
 	     let nrr = R.subst p v rr in
 	     if R.equal rr nrr then env, touched, neqs_to_up
@@ -470,6 +535,7 @@ module Make ( R : Sig.X ) = struct
       else
 	let env, tch, neqs_to_up = MapR.fold
 	  (fun r (rr,ex) (env,tch,neqs_to_up) ->
+	    !Options.thread_yield ();
 	     let nrr, ex_nrr = normal_form env rr in
 	     if R.equal nrr rr then env, tch, neqs_to_up
 	     else 
@@ -487,6 +553,7 @@ module Make ( R : Sig.X ) = struct
 	  
     let apply_sigma eqs env tch ((p, v, dep) as sigma) = 
       let env = init_leaf env p in
+      let env = init_leaf env v in
       let env = apply_sigma_ac eqs env sigma in
       let env, touched = apply_sigma_uf env sigma in 
       up_uf_rs dep env ((p, touched, v) :: tch)
@@ -494,16 +561,18 @@ module Make ( R : Sig.X ) = struct
   end
     
   let add env t = 
-    if rules = 3 then fprintf fmt "[rule] TR-UFX-Add@.";
+    if rules () = 3 then fprintf fmt "[rule] TR-UFX-Add@.";
     if MapT.mem t env.make then env, [] else Env.init_term env t
 
   let ac_solve eqs dep (env, tch) (p, v) = 
     (* pourquoi recuperer le representant de rv? r = rv d'apres testopt *)
-    if debug_uf then 
+    if debug_uf () then 
       printf "[uf] ac-solve: %a |-> %a %a@." R.print p R.print v Ex.print dep;
-    assert ( let rp, _ = Env.find_or_normal_form env p in R.equal p rp);
+    assert (not !Preoptions.enable_assertions ||
+              let rp, _ = Env.find_or_normal_form env p in R.equal p rp);
     let rv, ex_rv = Env.find_or_normal_form env v in
-    assert ( let rv, _ = Env.find_or_normal_form env v in R.equal v rv);
+    assert (not !Preoptions.enable_assertions ||
+              let rv, _ = Env.find_or_normal_form env v in R.equal v rv);
     let dep = Ex.union ex_rv dep in
     Env.apply_sigma eqs env tch (p, rv, dep)
 
@@ -511,11 +580,11 @@ module Make ( R : Sig.X ) = struct
     let rr1, ex_r1 = Env.find_or_normal_form env r1 in
     let rr2, ex_r2 = Env.find_or_normal_form env r2 in
     let dep = Ex.union dep (Ex.union ex_r1 ex_r2) in
-    if debug_uf then 
+    if debug_uf () then 
       printf "[uf] x-solve: %a = %a %a@."
 	R.print rr1 R.print rr2 Ex.print dep;
     if R.equal rr1 rr2 then begin
-      if rules = 3 then fprintf fmt "[rule] TR-CCX-Remove@.";
+      if rules () = 3 then fprintf fmt "[rule] TR-CCX-Remove@.";
       [] (* Remove rule *)
     end
     else 
@@ -523,24 +592,24 @@ module Make ( R : Sig.X ) = struct
 	ignore (Env.update_neqs rr1 rr2 dep env);
         try R.solve rr1 rr2 
 	with Unsolvable ->
-	  if rules = 3 then fprintf fmt "[rule] TR-CCX-Congruence-Conflict@.";
-	  raise (Inconsistent dep)
+	  if rules () = 3 then fprintf fmt "[rule] TR-CCX-Congruence-Conflict@.";
+	  raise (Inconsistent (dep, cl_extract env))
       end
         
   let rec ac_x eqs env tch = 
     if Queue.is_empty eqs then env, tch
     else 
       let r1, r2, dep = Queue.pop eqs in
-      if debug_uf then 
+      if debug_uf () then 
 	printf "[uf] ac(x): delta (%a) = delta (%a)@." 
 	  R.print r1 R.print r2;
       let sbs = x_solve env r1 r2 dep in
       let env, tch = List.fold_left (ac_solve eqs dep) (env, tch) sbs in
-      if debug_uf then Print.all fmt env;
+      if debug_uf () then Print.all fmt env;
       ac_x eqs env tch
       
   let union env r1 r2 dep =
-    if rules = 3 then fprintf fmt "[rule] TR-UFX-Union@.";
+    if rules () = 3 then fprintf fmt "[rule] TR-UFX-Union@.";
     let equations = Queue.create () in 
     Queue.push (r1,r2, dep) equations;
     ac_x equations env []
@@ -548,15 +617,16 @@ module Make ( R : Sig.X ) = struct
   let rec distinct env rl dep =
     Print.all fmt env;
     let d = Lit.make (Literal.Distinct (false,rl)) in
-    if debug_uf then fprintf fmt "[uf] distinct %a@." Lit.print d;
+    if debug_uf () then fprintf fmt "[uf] distinct %a@." Lit.print d;
     let env, _, newds = 
       List.fold_left
 	(fun (env, mapr, newds) r -> 
+	  !Options.thread_yield ();
 	   let rr, ex = Env.find_or_normal_form env r in 
 	   try
 	     let exr = MapR.find rr mapr in
-	     if rules = 3 then fprintf fmt "[rule] TR-CCX-Distinct-Conflict@.";
-	     raise (Inconsistent (Ex.union ex exr))
+	     if rules () = 3 then fprintf fmt "[rule] TR-CCX-Distinct-Conflict@.";
+	     raise (Inconsistent ((Ex.union ex exr), cl_extract env))
 	   with Not_found ->
 	     let uex = Ex.union ex dep in
 	     let mdis = 
@@ -585,9 +655,9 @@ module Make ( R : Sig.X ) = struct
 			    else
 			      distinct env [a; b] ex
 			| []  -> 
-			  if rules = 3 then
+			  if rules () = 3 then
 			    fprintf fmt "[rule] TR-CCX-Distinct-Conflict@.";
-			  raise (Inconsistent ex) 
+			  raise (Inconsistent (ex, cl_extract env)) 
 			| _   -> env
 		      with Unsolvable -> env) mapr env)
       env newds
@@ -596,40 +666,100 @@ module Make ( R : Sig.X ) = struct
   let are_equal env t1 t2 = 
     let r1, ex_r1 = Env.lookup_by_t t1 env in
     let r2, ex_r2 = Env.lookup_by_t t2 env in
-    if R.equal r1 r2 then Yes(Ex.union ex_r1 ex_r2) else No
+    if R.equal r1 r2 then
+      Yes (Ex.union ex_r1 ex_r2, cl_extract env)
+    else No
 
   let are_distinct env t1 t2 = 
-    if debug_uf then
+    if debug_uf () then
       printf " [uf] are_distinct %a %a @." T.print t1 T.print t2; 
     let r1, ex_r1 = Env.lookup_by_t t1 env in
     let r2, ex_r2 = Env.lookup_by_t t2 env in
     try
       ignore (union env r1 r2 (Ex.union ex_r1 ex_r2));
       No
-    with Inconsistent ex -> Yes(ex)
+    with Inconsistent (ex, classes) -> Yes (ex, classes)
 
-  let already_distinct env lr = 
+  let already_distinct env lr =
     let d = Lit.make (Literal.Distinct (false,lr)) in
-    try 
-      List.iter (fun r -> 
-	let mdis = MapR.find r env.neqs in
-	ignore (MapL.find d mdis)
+    try
+      List.iter (fun r ->
+  	let mdis = MapR.find r env.neqs in
+  	ignore (MapL.find d mdis)
       ) lr;
       true
     with Not_found -> false
 
+  (* let already_distinct env lr = *)
+  (*   let _, common = *)
+  (*     List.fold_left (fun (first, acc) r -> *)
+  (* 	try  *)
+  (* 	  let dis = *)
+  (* 	    MapL.fold (fun d _ acc -> Lit.Set.add d acc)  *)
+  (* 	      (MapR.find r env.neqs) Lit.Set.empty  *)
+  (* 	  in *)
+  (* 	  if first then false, dis else false, Lit.Set.inter acc dis *)
+  (* 	with Not_found -> false, Lit.Set.empty *)
+  (*     ) (true, Lit.Set.empty) lr *)
+  (*   in not (Lit.Set.is_empty common) *)
+
   let class_of env t = 
     try 
       let rt, _ = MapR.find (MapT.find t env.make) env.repr in
       SetT.elements (MapR.find rt env.classes)
     with Not_found -> [t]
-      
+
+
+  let mapt_choose m =
+    let r = ref None in
+    (try 
+       MapT.iter (fun x rx -> 
+	 r := Some (x, rx); raise Exit
+       ) m 
+     with Exit -> ());
+    match !r with Some b -> b | _ -> raise Not_found
+
+  let model env =
+    let eqs =
+      MapR.fold (fun r cl acc ->
+    	let l, to_rel =
+    	  List.fold_left (fun (l, to_rel) t ->
+    	    let rt = MapT.find t env.make in
+    	    if complete_model () || T.is_in_model t then
+    	      if rt <> r then t::l, (t,rt)::to_rel
+    	      else l, (t,rt)::to_rel
+    	    else l, to_rel
+    	  ) ([], []) (SetT.elements cl) in
+    	(r, l, to_rel)::acc
+      ) env.classes []
+    in
+    let rec extract_neqs acc makes =
+      try
+	let x, rx = mapt_choose makes in
+	let makes = MapT.remove x makes in
+	let acc =
+	  if complete_model () || T.is_in_model x then
+	    MapT.fold (fun y ry acc ->
+	      if (complete_model () || T.is_in_model y)
+		&& (already_distinct env [rx; ry] 
+		    || already_distinct env [ry; rx])
+	      then [y; x]::acc
+	      else acc
+	    ) makes acc
+	  else acc
+	in extract_neqs acc makes
+      with Not_found -> acc
+    in
+    let neqs = extract_neqs [] env.make in
+    eqs, neqs
+
+
   let find env t = 
-    if rules = 3 then fprintf fmt "[rule] TR-UFX-Find@.";
+    if rules () = 3 then fprintf fmt "[rule] TR-UFX-Find@.";
     Env.lookup_by_t t env
 
   let find_r = 
-    if rules = 3 then fprintf fmt "[rule] TR-UFX-Find@.";
+    if rules () = 3 then fprintf fmt "[rule] TR-UFX-Find@.";
     Env.find_or_normal_form
 
   let print = Print.all 
