@@ -23,12 +23,68 @@ open Lexing
 open Format
 open Options
 
+open Gui_session
+
 let indent_size = 4
-let monospace_font = Pango.Font.from_string "mono"
+let monospace_font = Pango.Font.from_string "monospace"
 let general_font = Pango.Font.from_string "sans"
 
 type sbuffer = GSourceView2.source_buffer
 
+type error_model = {
+  mutable some : bool;
+  rcols : GTree.column_list;
+  rcol_icon : GtkStock.id GTree.column;
+  rcol_desc : String.t GTree.column;
+  rcol_line : int GTree.column;
+  rcol_type : int GTree.column;
+  rcol_color : String.t GTree.column;
+  rstore : GTree.list_store;
+}
+
+type inst_model = {
+  h : (int, Gtk.tree_iter option ref * int ref * string * int ref) Hashtbl.t;
+  mutable max : int;
+  icols : GTree.column_list;
+  icol_icon : GtkStock.id GTree.column;
+  icol_desc : String.t GTree.column;
+  icol_number : int GTree.column;
+  icol_limit : String.t GTree.column;
+  icol_tag : int GTree.column;
+  istore : GTree.list_store;
+}
+
+type timers_model = {
+  timers : Timers.t;
+
+  label_sat : GMisc.label;
+  label_match : GMisc.label;
+  label_cc : GMisc.label;
+  label_arith : GMisc.label;
+  label_arrays : GMisc.label;
+  label_sum : GMisc.label;
+  label_records : GMisc.label;
+  label_ac : GMisc.label;
+
+  tl_sat : GMisc.label;
+  tl_match : GMisc.label;
+  tl_cc : GMisc.label;
+  tl_arith : GMisc.label;
+  tl_arrays : GMisc.label;
+  tl_sum : GMisc.label;
+  tl_records : GMisc.label;
+  tl_ac : GMisc.label;
+
+  pr_sat : GRange.progress_bar;
+  pr_match : GRange.progress_bar;
+  pr_cc : GRange.progress_bar;
+  pr_arith : GRange.progress_bar;
+  pr_arrays : GRange.progress_bar;
+  pr_sum : GRange.progress_bar;
+  pr_records : GRange.progress_bar;
+  pr_ac : GRange.progress_bar;
+}
+
 type 'a annoted =
     { mutable c : 'a; 
       mutable pruned : bool;
@@ -36,7 +92,9 @@ type 'a annoted =
       tag : GText.tag;
       ptag : GText.tag;
       id : int;
-      buf : sbuffer }
+      buf : sbuffer;
+      mutable line : int;
+    }
 
 type aterm = 
     { at_ty : Ty.t; at_desc : at_desc }
@@ -54,6 +112,7 @@ and at_desc =
   | ATlet of Symbols.t * aterm * aterm
   | ATdot of aterm * Hstring.t
   | ATrecord of (Hstring.t * aterm) list
+  | ATnamed of Hstring.t * aterm
 
       
       
@@ -88,9 +147,9 @@ and aform =
   | AFnamed of Hstring.t * aform annoted
 
 type atyped_decl = 
-  | AAxiom of loc * string * aform
+  | AAxiom of loc * string * inversion * aform
   | ARewriting of loc * string * ((aterm rwt_rule) annoted) list
-  | AGoal of loc * string * aform annoted
+  | AGoal of loc * goal_sort * string * aform annoted
   | ALogic of loc * string list * plogic_type
   | APredicate_def of loc * string * (string * ppure_type) list * aform
   | AFunction_def 
@@ -113,41 +172,84 @@ module MDep = Map.Make (
   end)
 
 
+module MTag = Map.Make (struct 
+  type t = GText.tag 
+  let compare t1 t2 = compare t1#get_oid t2#get_oid 
+end)
+
+
 type env = {
   buffer : sbuffer;
   inst_buffer : sbuffer;
+  errors : error_model;
+  insts : inst_model;
   st_ctx : GMisc.statusbar_context;
   mutable ast : (atyped_decl annoted * Why_typing.env) list;
   mutable ctrl : bool;
   mutable last_tag : GText.tag;
   mutable search_tags : GText.tag list;
-  mutable proof_tags : GText.tag list;
+  mutable proof_tags : int MTag.t;
   mutable proof_toptags : GText.tag list;
-  dep : (atyped_decl annoted list * atyped_decl annoted list) MDep.t
+  mutable start_select : int option;
+  mutable stop_select : int option;
+  dep : (atyped_decl annoted list * atyped_decl annoted list) MDep.t;
+  actions : action Stack.t;
+  saved_actions : action Stack.t;
+  resulting_ids : (string * int) list;
 }
 
-let create_env buf1 (buf2:sbuffer) st_ctx ast dep  =
+
+let create_env buf1 (buf2:sbuffer) errors insts st_ctx ast dep
+    actions resulting_ids=
   let titag = buf2#create_tag [`WEIGHT `BOLD; `UNDERLINE `SINGLE] in
   buf2#insert ~tags:[titag] "User instantiated axioms:\n\n";
   {
     buffer = buf1;
     inst_buffer = buf2;
+    errors = errors;
+    insts = insts;
     st_ctx = st_ctx;
     ast = ast;
     dep = dep;
     ctrl = false;
     last_tag = GText.tag ();
     search_tags = [];
-    proof_tags = [];
+    proof_tags = MTag.empty;
     proof_toptags = [];
+    start_select = None;
+    stop_select = None;
+    actions = actions;
+    saved_actions = Stack.copy actions;
+    resulting_ids = resulting_ids;
   }
 
+let create_replay_env buf1 errors insts ast actions resulting_ids =
+ {
+    buffer = buf1;
+    inst_buffer = GSourceView2.source_buffer ();
+    errors = errors;
+    insts = insts;
+    st_ctx = (GMisc.statusbar ())#new_context ~name:"";
+    ast = ast;
+    dep = MDep.empty;
+    ctrl = false;
+    last_tag = GText.tag ();
+    search_tags = [];
+    proof_tags = MTag.empty;
+    proof_toptags = [];
+    start_select = None;
+    stop_select = None;
+    actions = actions;
+    saved_actions = actions;
+    resulting_ids = resulting_ids;
+  }
 
 let tag (buffer:sbuffer) = buffer#create_tag []
 
 let new_annot (buffer:sbuffer) c id ptag =
   { c = c; pruned = false; tag = (tag buffer); 
-    ptag = ptag;id = id; polarity = true; buf = buffer }
+    ptag = ptag; id = id; polarity = true; buf = buffer;
+    line = buffer#line_count }
 
 
 let rec findin_aterm tag buffer { at_desc = at_desc } =
@@ -192,6 +294,7 @@ and findin_at_desc tag buffer = function
 	  else r
 	else r
     | ATrecord r -> let atl = List.map snd r in findin_aterm_list tag buffer atl
+    | ATnamed (_, t) -> findin_aterm tag buffer t
 	
 let findin_aatom tag buffer aa =
   match aa with
@@ -263,7 +366,7 @@ let findin_atyped_delc tag buffer (td, env) stop_decl =
   else if goodbuf && c > 0 then None
   else if stop_decl then Some (AD (td, env))
   else match td.c with
-    | AAxiom (_, _, af)
+    | AAxiom (_, _, _, af)
     | APredicate_def (_, _, _, af)
     | AFunction_def (_, _, _, _, af) ->
         let aaf = new_annot buffer af (-1) tag in
@@ -275,7 +378,7 @@ let findin_atyped_delc tag buffer (td, env) stop_decl =
 	    | Some _ -> acc
 	    | None -> findin_aterm_list tag buffer [rl; rr]
 	  ) rwtl None*)
-    | AGoal (_, _, aaf) ->
+    | AGoal (_, _, _, aaf) ->
 	let goodbuf =  aaf.buf#get_oid = buffer#get_oid in
 	let c = compare tag#priority aaf.tag#priority in
 	if goodbuf && c = 0 then Some (AF (aaf, None))
@@ -428,6 +531,7 @@ and print_tt_desc fmt = function
   | TTdot (t, c) ->
       fprintf fmt "%a.%s" print_tterm t (Hstring.view c)
   | TTrecord r -> fprintf fmt "{ %a }" (print_record ";") r
+  | TTnamed (lbl, t) -> fprintf fmt "%s:%a" (Hstring.view lbl) print_tterm t
 
 let print_tatom fmt a = match a.Why_ptree.c with
   | TAtrue -> fprintf fmt "true" 
@@ -494,10 +598,13 @@ let rec print_record_type fmt = function
       fprintf fmt "%s : %a; %a" c print_ppure_type ty print_record_type l
 
 let print_typed_decl fmt td = match td.Why_ptree.c with
-  | TAxiom (_, s, tf) -> fprintf fmt "axiom %s : %a" s print_tform tf
+  | TAxiom (_, s, true, tf) -> fprintf fmt "inversion %s : %a" s print_tform tf
+  | TAxiom (_, s, _, tf) -> fprintf fmt "axiom %s : %a" s print_tform tf
   | TRewriting (_, s, rwtl) -> 
     fprintf fmt "rewriting %s : %a" s print_rwt_list rwtl
-  | TGoal (_, s, tf) -> fprintf fmt "goal %s : %a" s print_tform tf
+  | TGoal (_, Thm, s, tf) -> fprintf fmt "goal %s : %a" s print_tform tf
+  | TGoal (_, Check, s, tf) -> fprintf fmt "check %s : %a" s print_tform tf
+  | TGoal (_, Cut, s, tf) -> fprintf fmt "cut %s : %a" s print_tform tf
   | TLogic (_, ls, ty) ->
       fprintf fmt "logic %a : %a" print_string_list ls print_plogic_type ty
   | TPredicate_def (_, p, spptl, tf) ->
@@ -607,6 +714,7 @@ and make_dep_at_desc d ex dep = function
 	     let dep = make_dep_string d ex dep (Hstring.view c) in
 	     make_dep_aterm d ex dep t)
 	  dep r
+    | ATnamed (_, t) -> make_dep_aterm d ex dep t
 
 let make_dep_aatom d ex dep = function
   | AAtrue | AAfalse -> dep
@@ -640,7 +748,7 @@ and make_dep_aaform d ex dep aaf = make_dep_aform d ex dep aaf.c
 
 let make_dep_atyped_decl dep d =
   match d.c with
-  | AAxiom (loc, s, af) -> make_dep_aform d [] dep af
+  | AAxiom (loc, s, _, af) -> make_dep_aform d [] dep af
   | ARewriting (loc, s, arwtl) ->
       List.fold_left
 	(fun dep r ->
@@ -649,7 +757,7 @@ let make_dep_atyped_decl dep d =
 	  let dep = make_dep_aterm d vars dep r.c.rwt_left in
 	  make_dep_aterm d vars dep r.c.rwt_right
 	) dep arwtl
-  | AGoal (loc, s, aaf) -> make_dep_aform d [] dep aaf.c
+  | AGoal (loc, _, s, aaf) -> make_dep_aform d [] dep aaf.c
   | ALogic (loc, ls, ty) -> MDep.add d ([], []) dep
   | APredicate_def (loc, p, spptl, af) ->
       let dep = MDep.add d ([], []) dep in
@@ -696,6 +804,7 @@ and of_tt_desc (buffer:sbuffer) = function
   | TTlet (s, t1, t2) -> ATlet (s, of_tterm buffer t1, of_tterm buffer t2)
   | TTdot (t, c) -> ATdot (of_tterm buffer t, c)
   | TTrecord r -> ATrecord (List.map (fun (c,t) -> (c, of_tterm buffer t)) r)
+  | TTnamed (lbl, t) -> ATnamed (lbl, of_tterm buffer t)
 
 let of_tatom (buffer:sbuffer) a = match a.Why_ptree.c with
   | TAtrue -> AAtrue
@@ -735,9 +844,9 @@ let rec of_quant_form (buffer:sbuffer)
     aqf_triggers = List.map (List.map (annot_of_tterm buffer )) trs;
     aqf_form = new_annot buffer (of_tform buffer tf) tf.Why_ptree.annot ptag}
 
-and annot_of_quant_form (buffer:sbuffer) qf =
+and annot_of_quant_form (buffer:sbuffer) qf id =
   let ptag = tag buffer in
-  new_annot buffer (of_quant_form buffer qf) (-1) ptag
+  new_annot buffer (of_quant_form buffer qf) id ptag
 
 and of_tform (buffer:sbuffer) f = match f.Why_ptree.c with
   | TFatom a -> AFatom (of_tatom buffer a)
@@ -747,8 +856,8 @@ and of_tform (buffer:sbuffer) f = match f.Why_ptree.c with
     if op = OPnot || op = OPimp then 
       change_polarity_aform (List.hd afl);
     AFop (of_oplogic buffer  op, afl)
-  | TFforall qf -> AFforall (annot_of_quant_form buffer qf)
-  | TFexists qf -> AFexists (annot_of_quant_form buffer qf)
+  | TFforall qf -> AFforall (annot_of_quant_form buffer qf f.Why_ptree.annot)
+  | TFexists qf -> AFexists (annot_of_quant_form buffer qf f.Why_ptree.annot)
   | TFlet (vs, s, t, tf) ->
       AFlet (vs, s, of_tterm buffer  t, annot_of_tform buffer tf)
   | TFnamed (n, tf) -> 
@@ -762,7 +871,7 @@ and annot_of_tform (buffer:sbuffer) t =
 let annot_of_typed_decl (buffer:sbuffer) td = 
   let ptag = tag buffer in
   let c = match td.Why_ptree.c with
-    | TAxiom (loc, s, tf) -> AAxiom (loc, s, of_tform buffer tf)
+    | TAxiom (loc, s, inv, tf) -> AAxiom (loc, s, inv, of_tform buffer tf)
     | TRewriting (loc, s, rwtl) ->
       let arwtl = List.map 
 	(fun rwt ->
@@ -773,9 +882,9 @@ let annot_of_typed_decl (buffer:sbuffer) td =
 	    td.Why_ptree.annot ptag
 	) rwtl in
       ARewriting (loc, s, arwtl)
-    | TGoal (loc, s, tf) ->
+    | TGoal (loc, gs, s, tf) ->
         let g = new_annot buffer (of_tform buffer tf) tf.Why_ptree.annot ptag in
-        AGoal (loc, s, g)
+        AGoal (loc, gs, s, g)
     | TLogic (loc, ls, ty) -> ALogic (loc, ls, ty)
     | TPredicate_def (loc, p, spptl, tf) ->
         APredicate_def (loc, p,  spptl, of_tform buffer  tf)
@@ -815,6 +924,7 @@ and to_tt_desc = function
     | ATlet (s, t1, t2) -> TTlet (s, to_tterm 0 t1, to_tterm 0 t2)
     | ATdot (t, c) -> TTdot (to_tterm 0 t, c)
     | ATrecord r -> TTrecord (List.map (fun (c, t) -> (c, to_tterm 0 t)) r)
+    | ATnamed (lbl, t) -> TTnamed (lbl, to_tterm 0 t)
 
 let to_tatom aa id = 
   let c = match aa with 
@@ -888,9 +998,9 @@ and from_aaform_list = function
 
 let to_typed_decl td =
   let c = match td.c with
-    | AAxiom (loc, s, af) -> 
+    | AAxiom (loc, s, inv, af) -> 
       let af = void_to_tform af td.id in
-      TAxiom (loc, s, af)
+      TAxiom (loc, s, inv, af)
     | ARewriting (loc, s, arwtl) -> 
       let rwtl = List.fold_left (fun rwtl ar ->
 	if ar.pruned then rwtl
@@ -899,7 +1009,7 @@ let to_typed_decl td =
 	       rwt_right = to_tterm ar.id ar.c.rwt_right}::rwtl
       ) [] arwtl in
       TRewriting (loc, s, rwtl)
-    | AGoal (loc, s, aaf) -> TGoal (loc, s, to_tform aaf)
+    | AGoal (loc, gs, s, aaf) -> TGoal (loc, gs, s, to_tform aaf)
     | ALogic (loc, ls, ty) -> TLogic (loc, ls, ty)
     | APredicate_def (loc, p, spptl, af) ->
       TPredicate_def (loc, p, spptl, void_to_tform af td.id)
@@ -936,9 +1046,11 @@ and add_aterm_list_at (buffer:sbuffer) tags iter sep =
 	add_aterm_list_at buffer tags iter sep l
 
 and add_aaterm_at (buffer:sbuffer) tags iter at =
+  at.line <- iter#line;
   add_aterm_at buffer (at.tag::at.ptag::tags) iter at.c
 
 and add_aaterm (buffer:sbuffer) tags at =
+  at.line <- buffer#line_count;
   add_aaterm_at buffer tags buffer#end_iter at
 
 and add_aaterm_list_at (buffer:sbuffer) tags iter sep =
@@ -1026,7 +1138,9 @@ and add_at_desc_at (buffer:sbuffer) tags iter at =
 	buffer#insert ~iter ~tags "{ ";
 	add_arecord_at buffer tags iter r;
 	buffer#insert ~iter ~tags " }"
-
+    | ATnamed (n, t) ->
+	buffer#insert ~iter ~tags (sprintf "%s: " (Hstring.view n));
+	add_aterm_at buffer tags iter t
 	
 	
 let add_aatom (buffer:sbuffer) indent tags aa =
@@ -1054,9 +1168,9 @@ let add_oplogic (buffer:sbuffer) indent tags op =
   | AOPnot -> buffer#insert ~tags "not "
   | AOPif at ->
       buffer#insert (String.make indent ' ');
-      buffer#insert ~tags "if  ";
+      buffer#insert ~tags "if ";
       add_aterm buffer tags at;
-      buffer#insert ~tags "  then"
+      buffer#insert ~tags " then "
   | AOPiff -> buffer#insert ~tags "<-> "
 
 let add_rwt (buffer:sbuffer) indent tags r =
@@ -1079,38 +1193,55 @@ let rec add_rwt_list (buffer:sbuffer) indent tags = function
     buffer#insert "\n";
     add_rwt_list buffer indent tags l
 
-let rec add_quant_form (buffer:sbuffer) indent tags qf =
+
+let add_empty_triggers_error ({rstore = rstore} as errors) (buffer:sbuffer) =
+  let row = rstore#append () in
+  rstore#set ~row ~column:errors.rcol_icon `DIALOG_WARNING;
+  rstore#set ~row ~column:errors.rcol_desc "Warning : Empty trigger, this lemma won't be instantiated.";
+  rstore#set ~row ~column:errors.rcol_color "red";
+  rstore#set ~row ~column:errors.rcol_type 1;
+  rstore#set ~row ~column:errors.rcol_line buffer#line_count;
+  errors.some <- true
+  
+
+let rec add_quant_form errors (buffer:sbuffer) indent tags qf =
   let {aqf_bvars = bv; aqf_upvars = uv; aqf_triggers = trs; aqf_form = aaf } =
     qf.c in
   fprintf str_formatter "%a " print_var_list bv;
   buffer#insert ~tags (flush_str_formatter ());
   let ntags = qf.tag::qf.ptag::tags in
   buffer#insert ~tags:ntags "[";
-  add_triggers buffer ntags trs;
+  add_triggers errors buffer ntags trs;
   buffer#insert ~tags:ntags "].";
-  ignore(buffer#create_source_mark ~category:"trigger" buffer#end_iter);
+  ignore(buffer#create_source_mark 
+	   ~category:(sprintf "trigger_%d" qf.id) buffer#end_iter);
   buffer#insert "\n";
   buffer#insert (String.make ((indent + 1)* indent_size) ' ');
-  add_aaform buffer (indent+1) tags aaf
+  add_aaform errors buffer (indent+1) tags aaf
+
+and add_triggers errors (buffer:sbuffer) tags triggers =
+  let rec add_triggers_aux  = function
+    | [] -> ()
+    | [atl] -> add_aaterm_list buffer tags "," atl
+    | atl::l -> 
+	add_aaterm_list buffer tags "," atl;
+	buffer#insert ~tags " | ";
+	add_triggers_aux l
+  in
+  if triggers = [] then add_empty_triggers_error errors buffer
+  else add_triggers_aux triggers
 
-and add_triggers (buffer:sbuffer) tags = function
-  | [] -> ()
-  | [atl] -> add_aaterm_list buffer tags "," atl
-  | atl::l -> 
-      add_aaterm_list buffer tags "," atl;
-      buffer#insert ~tags " | ";
-      add_triggers buffer tags l
       
-and add_aform (buffer:sbuffer) indent tags aform =
+and add_aform errors (buffer:sbuffer) indent tags aform =
   match aform with
   | AFatom a -> add_aatom buffer 0 tags a
-  | AFop (op, afl) -> add_aaform_list buffer indent tags op afl
+  | AFop (op, afl) -> add_aaform_list errors buffer indent tags op afl
   | AFforall qf ->
       buffer#insert ~tags "forall ";
-      add_quant_form buffer indent tags qf
+      add_quant_form errors buffer indent tags qf
   | AFexists qf ->
       buffer#insert ~tags "exists ";
-      add_quant_form buffer indent tags qf
+      add_quant_form errors buffer indent tags qf
   | AFlet (vs, s, t, aaf) ->
       buffer#insert ~tags 
 	(sprintf "let %s = " (Symbols.to_string s));
@@ -1118,32 +1249,32 @@ and add_aform (buffer:sbuffer) indent tags aform =
       buffer#insert ~tags " in";
       buffer#insert "\n";
       buffer#insert (String.make (indent * indent_size) ' ');
-      add_aaform buffer (indent) tags aaf
+      add_aaform errors buffer (indent) tags aaf
   | AFnamed (n, aaf) ->
       buffer#insert ~tags (sprintf "%s: " (Hstring.view n));
-      add_aform buffer indent tags aaf.c
+      add_aform errors buffer indent tags aaf.c
       
 
-and add_aaform_list (buffer:sbuffer) indent tags op l =
+and add_aaform_list errors (buffer:sbuffer) indent tags op l =
   if l = [] then ()
   else begin
     (* add_aaform buffer indent tags (List.hd l); *)
-    add_aaform_list_aux buffer indent tags op l
+    add_aaform_list_aux errors buffer indent tags op l
   end
    
-and add_aaform_list_aux (buffer:sbuffer) indent tags op =
+and add_aaform_list_aux errors (buffer:sbuffer) indent tags op =
   function
   | [] -> ()
   | [af] ->
       add_oplogic buffer indent tags op;
-      add_aaform buffer indent tags af
+      add_aaform errors buffer indent tags af
   | af1::af2::l ->
-      add_aaform buffer indent tags af1;
+      add_aaform errors buffer indent tags af1;
       buffer#insert "\n";
       buffer#insert (String.make (indent * indent_size) ' ');
       add_oplogic buffer indent tags op;
-      add_aaform buffer (indent+1) tags af2;
-      add_aaform_list buffer (indent+1) tags op l      
+      add_aaform errors buffer (indent+1) tags af2;
+      add_aaform_list errors buffer (indent+1) tags op l      
   (* | af::l -> *)
   (*     buffer#insert "\n"; *)
   (*     buffer#insert (String.make (indent * indent_size) ' '); *)
@@ -1152,38 +1283,45 @@ and add_aaform_list_aux (buffer:sbuffer) indent tags op =
   (*     add_aaform_list buffer (indent+1) tags op l *)
 
 
-and add_aaform (buffer:sbuffer) indent tags
-    {c = af; tag = tag; ptag = ptag} =
-  add_aform buffer indent (tag::ptag::tags) af
+and add_aaform errors (buffer:sbuffer) indent tags
+    ({c = af; tag = tag; ptag = ptag} as aaf) =
+  aaf.line <- buffer#line_count;
+  add_aform errors buffer indent (tag::ptag::tags) af
 
-let add_atyped_decl (buffer:sbuffer) d =
+let add_atyped_decl errors (buffer:sbuffer) d =
   match d.c with
-  | AAxiom (loc, s, af) ->
+  | AAxiom (loc, s, inv, af) ->
       let keyword = 
 	if String.length s > 0 && (s.[0] = '_'  || s.[0] = '@') 
-	then "hypothesis" else "axiom" in
+	then "hypothesis" else if inv then "inversion" else "axiom" in
       buffer#insert ~tags:[d.tag;d.ptag] (sprintf "%s %s :" keyword s);
       buffer#insert "\n";
+      d.line <- buffer#line_count;
       buffer#insert (String.make indent_size ' ');
-      add_aform buffer 1 [d.tag;d.ptag] af;
+      add_aform errors buffer 1 [d.tag;d.ptag] af;
       buffer#insert "\n\n"
 
   | ARewriting (loc, s, arwtl) ->
       buffer#insert ~tags:[d.tag;d.ptag] (sprintf "rewriting %s :" s);
       buffer#insert "\n";
+      d.line <- buffer#line_count;
       add_rwt_list buffer 1 [d.tag;d.ptag] arwtl;
       buffer#insert "\n\n"
 
-  | AGoal (loc, s, aaf) -> 
-      buffer#insert ~tags:[d.tag;d.ptag] (sprintf "goal %s :" s);
+  | AGoal (loc, gs, s, aaf) ->
+      let goal_str =
+	match gs with Thm -> "goal" | Check -> "check" | Cut -> "cut" in
+      buffer#insert ~tags:[d.tag;d.ptag] (sprintf "%s %s :" goal_str s);
       buffer#insert "\n";
+      d.line <- buffer#line_count;
       buffer#insert (String.make indent_size ' ');
-      add_aform buffer 1 [d.tag;d.ptag] aaf.c;
+      add_aform errors buffer 1 [d.tag;d.ptag] aaf.c;
       buffer#insert "\n\n"
 
   | ALogic (loc, ls, ty) ->
       fprintf str_formatter 
 	"logic %a : %a" print_string_list ls print_plogic_type ty;
+      d.line <- buffer#line_count;
       buffer#insert ~tags:[d.tag;d.ptag] (flush_str_formatter());
       buffer#insert "\n\n"
 
@@ -1191,8 +1329,9 @@ let add_atyped_decl (buffer:sbuffer) d =
       fprintf str_formatter "predicate %s %a =" p print_pred_type_list spptl;
       buffer#insert ~tags:[d.tag;d.ptag] (flush_str_formatter());
       buffer#insert "\n";
+      d.line <- buffer#line_count;
       buffer#insert (String.make indent_size ' ');
-      add_aform buffer 1 [d.tag;d.ptag] af;
+      add_aform errors buffer 1 [d.tag;d.ptag] af;
       buffer#insert "\n\n"
 
   | AFunction_def (loc, f, spptl, ty, af) ->
@@ -1200,32 +1339,34 @@ let add_atyped_decl (buffer:sbuffer) d =
 	print_string_ppure_type_list spptl print_ppure_type ty;
       buffer#insert ~tags:[d.tag;d.ptag] (flush_str_formatter());
       buffer#insert "\n";
+      d.line <- buffer#line_count;
       buffer#insert (String.make indent_size ' ');
-      add_aform buffer 1 [d.tag;d.ptag] af;
+      add_aform errors buffer 1 [d.tag;d.ptag] af;
       buffer#insert "\n\n"
       
   | ATypeDecl (loc, ls, s, Abstract) -> 
       fprintf str_formatter "type %a %s" print_astring_list ls s;
+      d.line <- buffer#line_count;
       buffer#insert ~tags:[d.tag;d.ptag] (flush_str_formatter());
       buffer#insert "\n\n"
       
   | ATypeDecl (loc, ls, s, Enum lc) -> 
       fprintf str_formatter "type %a %s = %a"
 	print_astring_list ls s (print_string_sep " | ") lc;
+      d.line <- buffer#line_count;
       buffer#insert ~tags:[d.tag;d.ptag] (flush_str_formatter());
       buffer#insert "\n\n"
 
   | ATypeDecl (loc, ls, s, Record rt) -> 
       fprintf str_formatter "type %a %s = { %a }"
 	print_astring_list ls s	print_record_type rt;
+      d.line <- buffer#line_count;
       buffer#insert ~tags:[d.tag;d.ptag] (flush_str_formatter());
       buffer#insert "\n\n"
 
 
-let add_to_buffer (buffer:sbuffer) annoted_ast =
-  List.iter (fun (t, _) -> add_atyped_decl buffer t) annoted_ast
-
-
+let add_to_buffer errors (buffer:sbuffer) annoted_ast =
+  List.iter (fun (t, _) -> add_atyped_decl errors buffer t) annoted_ast
 
 
 
@@ -1242,7 +1383,7 @@ let rec isin_aterm sl { at_desc = at_desc } =
     | ATinfix (t1, _, t2) | ATget (t1,t2)
     | ATconcat (t1, t2) | ATlet (_, t1, t2) ->
       isin_aterm sl t1 || isin_aterm sl t2
-    | ATdot (t, _ ) | ATprefix (_,t) -> isin_aterm sl t
+    | ATdot (t, _ ) | ATprefix (_,t) | ATnamed (_, t) -> isin_aterm sl t
     | ATset (t1, t2, t3) | ATextract (t1, t2, t3)  ->
       isin_aterm sl t1 || isin_aterm sl t2 || isin_aterm sl t3 
     | ATrecord rt -> let atl = List.map snd rt in isin_aterm_list sl atl
@@ -1263,7 +1404,7 @@ and findtags_aaterm sl aat acc =
     | ATinfix (t1, _, t2) | ATget (t1,t2)
     | ATconcat (t1, t2) | ATlet (_, t1, t2) ->
 	if isin_aterm sl t1 || isin_aterm sl t2 then aat.tag::acc else acc
-    | ATdot (t, _) | ATprefix (_,t) ->
+    | ATdot (t, _) | ATprefix (_,t) | ATnamed (_, t) ->
 	if isin_aterm sl t then aat.tag::acc else acc
     | ATset (t1, t2, t3) | ATextract (t1, t2, t3)  ->
 	if isin_aterm sl t1 || isin_aterm sl t2 || isin_aterm sl t3 
@@ -1329,12 +1470,12 @@ and findtags_aaform sl aaf acc =
 
 let findtags_atyped_delc sl td acc =
   match td.c with
-    | AAxiom (_, _, af)
+    | AAxiom (_, _, _, af)
     | APredicate_def (_, _, _, af)
     | AFunction_def (_, _, _, _, af) ->
 	findtags_aform sl af acc
     | ARewriting (_, _, rwtl) -> acc
-    | AGoal (_, _, aaf) ->
+    | AGoal (_, _, _, aaf) ->
 	findtags_aform sl aaf.c acc
     | ALogic _
     | ATypeDecl _ -> acc
@@ -1366,7 +1507,7 @@ let rec listsymbols at acc =
     | ATinfix (t1, _, t2) | ATget (t1,t2)
     | ATconcat (t1, t2) | ATlet (_, t1, t2) ->
 	listsymbols t1 (listsymbols t2 acc)
-    | ATdot (t, _) | ATprefix (_,t) -> listsymbols t acc
+    | ATdot (t, _) | ATprefix (_,t) | ATnamed (_, t) -> listsymbols t acc
     | ATset (t1, t2, t3) | ATextract (t1, t2, t3)  ->
 	listsymbols t1 (listsymbols t2 (listsymbols t3 acc))
     | ATrecord r ->
@@ -1386,29 +1527,45 @@ let findtags_dep at l =
   let sl = listsymbols at [] in
   List.fold_left (fun acc (td, _) -> findtags_atyped_delc_dep sl td acc) [] l
   
-let rec findproof_aform ids af acc found =
+let rec findproof_aform ids af acc depth found =
   match af with
     | AFatom at -> acc, found
+    | AFop ((AOPand), aafl) ->
+      List.fold_left
+    	(fun (acc, found) aaf -> findproof_aaform ids aaf acc depth found)
+    	(acc,found) aafl
     | AFop (_, aafl) ->
       List.fold_left
-	(fun (acc, found) aaf -> findproof_aaform ids aaf acc found)
+	(fun (acc, found) aaf -> 
+	  findproof_aaform ids aaf acc depth found)
 	(acc,found) aafl
     | AFforall aaqf | AFexists aaqf ->
-      let acc, found = 
-	if List.mem aaqf.id ids then aaqf.ptag::acc, true 
-	else acc, found in
-      findproof_aaform ids aaqf.c.aqf_form acc found
+      let acc, found =
+	try 
+	  let m = Explanation.MI.find aaqf.id ids in
+	  MTag.add aaqf.ptag m acc, true
+	with Not_found -> acc, found
+      in
+      findproof_aaform ids aaqf.c.aqf_form acc depth found
     | AFlet (_,_,_, aaf) | AFnamed (_, aaf) ->
-      findproof_aaform ids aaf acc found
+      findproof_aaform ids aaf acc depth found
 
-and findproof_aaform ids aaf acc found =
-  let acc, found = 
-    if List.mem aaf.id ids then aaf.ptag::acc, true 
-    else acc, found in
-  findproof_aform ids aaf.c acc found
+and findproof_aaform ids aaf acc depth found =
+  let acc, found =
+    try 
+      let m = Explanation.MI.find aaf.id ids in
+      MTag.add aaf.ptag m acc, true
+    with Not_found -> acc, found
+  in
+  findproof_aform ids aaf.c acc (depth) found
 
 let findproof_atyped_decl ids td (ax,acc) =
-  let acc = if List.mem td.id ids then td.ptag::acc else acc in
+  let acc =
+    try 
+      let m = Explanation.MI.find td.id ids in
+      MTag.add td.ptag m acc
+    with Not_found -> acc
+  in
   match td.c with
     | ARewriting (_,_, arwtl) -> assert false
 
@@ -1416,17 +1573,146 @@ let findproof_atyped_decl ids td (ax,acc) =
 
     | APredicate_def (_,_,_, af) 
     | AFunction_def (_,_,_,_, af) 
-    | AAxiom (_, _, af) -> 
-      let acc, found = findproof_aform ids af acc false in
+    | AAxiom (_, _, _, af) -> 
+      let acc, found = findproof_aform ids af acc 1 false in
       if found then td.ptag::ax, acc else ax,acc
 
-    | AGoal (_,_, aaf) ->
-      let acc, found = findproof_aaform ids aaf acc false in
+    | AGoal (_,_,_, aaf) ->
+      let acc, found = findproof_aaform ids aaf acc 1 false in
       if found then td.ptag::ax, acc else ax,acc
 
 let findtags_proof expl l =
-  match Explanation.ids_of expl with
-    | None -> [],[]
-    | Some ids -> 
-      List.fold_left (fun acc (td, _) ->
-	findproof_atyped_decl ids td acc) ([],[]) l
+  let ids = Explanation.literals_ids_of expl in
+  List.fold_left (fun acc (td, _) ->
+    findproof_atyped_decl ids td acc) ([], MTag.empty) l
+
+
+exception FoundLine of int * GText.tag
+
+let rec find_line_id_aform id af =
+  match af with    
+    | AFatom at -> ()
+    | AFop (_, aafl) ->
+      List.iter (find_line_id_aaform id) aafl
+    | AFforall aaqf | AFexists aaqf ->
+	if aaqf.id = id then raise (FoundLine (aaqf.line, aaqf.tag))
+	else find_line_id_aaform id aaqf.c.aqf_form
+    | AFlet (_,_,_, aaf) | AFnamed (_, aaf) ->
+      find_line_id_aaform id aaf
+
+and find_line_id_aaform id aaf =
+  if aaf.id = id then raise (FoundLine (aaf.line, aaf.tag))
+  else find_line_id_aform id aaf.c
+
+let find_line_id_atyped_decl id td =
+  if td.id < id then ()
+  else if td.id = id then raise (FoundLine (td.line, td.tag))
+  else match td.c with
+    | ARewriting (_,_, _) 
+    | ALogic _ | ATypeDecl _  -> ()
+
+    | APredicate_def (_,_,_, af) 
+    | AFunction_def (_,_,_,_, af) 
+    | AAxiom (_, _, _, af) ->
+	find_line_id_aform id af
+
+    | AGoal (_,_,_, aaf) ->
+	find_line_id_aaform id aaf
+
+let find_line id l =
+  try
+    List.iter (fun (d, _) -> find_line_id_atyped_decl id d) l;
+    raise Not_found
+  with FoundLine (line, tag) -> line, tag
+
+
+
+exception Foundannot of annoted_node
+
+
+
+let findbyid_aaterm id aat =
+  if aat.id = id then raise (Foundannot (AT aat))
+
+(*   else findbyid_atdesc id aat.c.at_desc *)
+
+(* and findbyid_atdesc id = function *)
+(*   | ATconst _ | ATvar _ -> () *)
+(*   | ATapp (s, atl) -> List.iter (findbyid_aaterm id) atl *)
+(*   | ATinfix (t1, _, t2) | ATget (t1,t2) *)
+(*   | ATconcat (t1, t2) | ATlet (_, t1, t2) -> *)
+(*     findbyid_aaterm id t1; *)
+(*     findbyid_aaterm id t2 *)
+(*   | ATdot (t, _) | ATprefix (_,t) | ATnamed (_, t) ->  findbyid_aaterm id t *)
+(*   | ATset (t1, t2, t3) | ATextract (t1, t2, t3)  -> *)
+(*     findbyid_aaterm id t1; *)
+(*     findbyid_aaterm id t2; *)
+(*     findbyid_aaterm id t3 *)
+(*   | ATrecord r ->  *)
+(*     let atl = List.map snd r in List.iter (findbyid_aaterm id) atl     *)
+
+let findbyid_aatom id = function
+  | AAtrue
+  | AAfalse -> ()
+
+  | AAeq atl
+  | AAneq atl
+  | AAdistinct atl
+  | AAle atl
+  | AAlt atl
+  | AAbuilt (_, atl) -> List.iter (findbyid_aaterm id) atl
+
+  | AApred at -> ()
+
+let rec findbyid_aform id af =
+  match af with    
+    | AFatom aat -> findbyid_aatom id aat
+    | AFop (_, aafl) ->
+      List.iter (findbyid_aaform id) aafl
+    | AFforall aaqf | AFexists aaqf ->
+        List.iter (List.iter (findbyid_aaterm id)) aaqf.c.aqf_triggers;
+	if aaqf.id = id then raise (Foundannot (QF aaqf))
+	else findbyid_aaform id aaqf.c.aqf_form
+    | AFlet (_,_,_, aaf) | AFnamed (_, aaf) ->
+      findbyid_aaform id aaf
+
+and findbyid_aaform id aaf =
+  if aaf.id = id then raise (Foundannot (AF (aaf, None)))
+  else findbyid_aform id aaf.c
+
+let findbyid_atyped_decl  stop_decl id (td, tyenv) =
+  if td.id < id then ()
+  else if td.id = id then raise (Foundannot (AD (td, tyenv)))
+  else if stop_decl then raise (Foundannot (AD (td, tyenv)))
+  else match td.c with
+    | ARewriting (_,_, _) 
+    | ALogic _ | ATypeDecl _  -> ()
+
+    | APredicate_def (_,_,_, af) 
+    | AFunction_def (_,_,_,_, af) 
+    | AAxiom (_, _, _, af) ->
+	findbyid_aform id af
+
+    | AGoal (_,_,_, aaf) ->
+	findbyid_aaform id aaf
+
+let findbyid_aux stop_decl id l =
+  try
+    List.iter (findbyid_atyped_decl stop_decl id) l;
+    raise Not_found
+  with Foundannot a -> a
+
+let findbyid = findbyid_aux false
+
+let findbyid_decl = findbyid_aux true
+
+let compute_resulting_ids =
+  List.fold_left (fun acc (td, _) -> match td.c with
+    | ARewriting (_,_, _) -> acc
+    | ALogic (_, names, _) -> (List.map (fun n -> n, td.id) names)@acc
+    | ATypeDecl (_, _, name, _)
+    | APredicate_def (_, name, _, _) 
+    | AFunction_def (_, name, _, _, _) 
+    | AAxiom (_, name, _, _)
+    | AGoal (_,_, name, _) -> (name, td.id)::acc)
+    []
