@@ -18,9 +18,64 @@
 (**************************************************************************)
 
 open Why_ptree
+open Gui_session
 
 type sbuffer = GSourceView2.source_buffer
 
+type error_model = {
+  mutable some : bool;
+  rcols : GTree.column_list;
+  rcol_icon : GtkStock.id GTree.column;
+  rcol_desc : String.t GTree.column;
+  rcol_line : int GTree.column;
+  rcol_type : int GTree.column;
+  rcol_color : String.t GTree.column;
+  rstore : GTree.list_store;
+}
+
+type inst_model = {
+  h : (int, Gtk.tree_iter option ref * int ref * string * int ref) Hashtbl.t;
+  mutable max : int;
+  icols : GTree.column_list;
+  icol_icon : GtkStock.id GTree.column;
+  icol_desc : String.t GTree.column;
+  icol_number : int GTree.column;
+  icol_limit : String.t GTree.column;
+  icol_tag : int GTree.column;
+  istore : GTree.list_store;
+}
+
+type timers_model = {
+  timers : Timers.t;
+
+  label_sat : GMisc.label;
+  label_match : GMisc.label;
+  label_cc : GMisc.label;
+  label_arith : GMisc.label;
+  label_arrays : GMisc.label;
+  label_sum : GMisc.label;
+  label_records : GMisc.label;
+  label_ac : GMisc.label;
+
+  tl_sat : GMisc.label;
+  tl_match : GMisc.label;
+  tl_cc : GMisc.label;
+  tl_arith : GMisc.label;
+  tl_arrays : GMisc.label;
+  tl_sum : GMisc.label;
+  tl_records : GMisc.label;
+  tl_ac : GMisc.label;
+
+  pr_sat : GRange.progress_bar;
+  pr_match : GRange.progress_bar;
+  pr_cc : GRange.progress_bar;
+  pr_arith : GRange.progress_bar;
+  pr_arrays : GRange.progress_bar;
+  pr_sum : GRange.progress_bar;
+  pr_records : GRange.progress_bar;
+  pr_ac : GRange.progress_bar;
+}
+
 type 'a annoted =
     { mutable c : 'a; 
       mutable pruned : bool;
@@ -28,7 +83,9 @@ type 'a annoted =
       tag : GText.tag;
       ptag : GText.tag;
       id : int;
-      buf : sbuffer }
+      buf : sbuffer;
+      mutable line : int;
+ }
 
 type aterm = 
     { at_ty : Ty.t; at_desc : at_desc }
@@ -46,6 +103,7 @@ and at_desc =
   | ATlet of Symbols.t * aterm * aterm
   | ATdot of aterm * Hstring.t
   | ATrecord of (Hstring.t * aterm) list
+  | ATnamed of Hstring.t * aterm
       
 
 type aatom = 
@@ -78,9 +136,9 @@ and aform =
   | AFnamed of Hstring.t * aform annoted
 
 type atyped_decl = 
-  | AAxiom of loc * string * aform
+  | AAxiom of loc * string * inversion * aform
   | ARewriting of loc * string * ((aterm rwt_rule) annoted) list
-  | AGoal of loc * string * aform annoted
+  | AGoal of loc * goal_sort * string * aform annoted
   | ALogic of loc * string list * plogic_type
   | APredicate_def of loc * string * (string * ppure_type) list * aform
   | AFunction_def 
@@ -95,21 +153,29 @@ type annoted_node =
   | QF of aquant_form annoted
 
 
-module MDep : (Map.S with type key = atyped_decl annoted)
+module MDep : Map.S with type key = atyped_decl annoted
 
+module MTag : Map.S with type key = GText.tag
 
 
 type env = {
   buffer : sbuffer;
   inst_buffer : sbuffer;
+  errors : error_model;
+  insts : inst_model;
   st_ctx : GMisc.statusbar_context;
-  mutable ast : (atyped_decl annoted * Why_typing.env)list;
+  mutable ast : (atyped_decl annoted * Why_typing.env) list;
   mutable ctrl : bool;
   mutable last_tag : GText.tag;
   mutable search_tags : GText.tag list;
-  mutable proof_tags : GText.tag list;
+  mutable proof_tags : int MTag.t;
   mutable proof_toptags : GText.tag list;
-  dep : (atyped_decl annoted list * atyped_decl annoted list) MDep.t
+  mutable start_select : int option;
+  mutable stop_select : int option;
+  dep : (atyped_decl annoted list * atyped_decl annoted list) MDep.t;
+  actions : action Stack.t;
+  saved_actions : action Stack.t;
+  resulting_ids : (string * int) list;
 }
 
 val indent_size : int
@@ -119,9 +185,22 @@ val general_font : Pango.font_description
 val create_env :
   sbuffer -> 
   sbuffer ->
+  error_model ->
+  inst_model ->
   GMisc.statusbar_context ->
   (atyped_decl annoted * Why_typing.env) list -> 
   (atyped_decl annoted list * atyped_decl annoted list) MDep.t ->
+  action Stack.t ->
+  (string * int) list ->
+  env
+
+val create_replay_env :
+  sbuffer -> 
+  error_model ->
+  inst_model ->
+  (atyped_decl annoted * Why_typing.env) list ->
+  action Stack.t ->
+  (string * int) list ->
   env
 
 val find : 
@@ -158,7 +237,7 @@ val annot_of_tterm :
 val add_aaterm_list_at : sbuffer ->
   GText.tag list -> GText.iter -> string -> aterm annoted list -> unit
 
-val add_aaform : sbuffer -> int -> GText.tag list ->
+val add_aaform : error_model -> sbuffer -> int -> GText.tag list ->
   aform annoted -> unit
 
 val to_ast : 
@@ -166,7 +245,7 @@ val to_ast :
   (int tdecl, int) Why_ptree.annoted list
 
 val add_to_buffer : 
-  sbuffer -> (atyped_decl annoted * Why_typing.env) list -> unit
+  error_model -> sbuffer -> (atyped_decl annoted * Why_typing.env) list -> unit
 
 val print_typed_decl_list  :
   Format.formatter -> (int tdecl, int) Why_ptree.annoted list -> unit
@@ -179,4 +258,16 @@ val findtags_dep :
 
 val findtags_proof : 
   Explanation.t -> (atyped_decl annoted * Why_typing.env) list ->
-  GText.tag list * GText.tag list 
+  GText.tag list * int MTag.t
+
+val find_line : 
+  int -> (atyped_decl annoted * 'a) list -> int * GText.tag
+
+val findbyid : 
+  int -> (atyped_decl annoted * Why_typing.env) list -> annoted_node
+
+val findbyid_decl : 
+  int -> (atyped_decl annoted * Why_typing.env) list -> annoted_node
+
+val compute_resulting_ids : 
+  (atyped_decl annoted * Why_typing.env) list -> (string * int) list
