@@ -22,43 +22,44 @@ open Why_annoted
 open Lexing
 open Format
 open Options
+open Gui_session
 
-    
-let rec prune r t dep =
+let prune ?(register=true) env r =
   r.pruned <- true;
-  t#set_property (`FOREGROUND "light gray");
-  let deps = match find_tag_inversedeps dep t  with 
+  if register then save env.actions (Prune r.id);
+  r.tag#set_property (`FOREGROUND "light gray")
+
+let incorrect_prune ?(register=true) env r =
+  r.pruned <- true;
+  if register then save env.actions (IncorrectPrune r.id);
+  r.tag#set_property (`FOREGROUND "tomato")
+
+let unprune ?(register=true) env r =
+  r.pruned <- false;
+  if register then save env.actions (Unprune r.id);
+  r.tag#set_property (`FOREGROUND_SET false)
+
+let rec prune_dep env r =
+  prune env r;
+  let deps = match find_tag_inversedeps env.dep r.tag  with 
     | None -> []
     | Some d -> d in
-  List.iter (fun d -> prune d d.tag dep) deps
+  List.iter (fun d -> prune_dep env d) deps
 
-let rec unprune r t dep =
-  r.pruned <- false;
-  t#set_property (`FOREGROUND_SET false);
-  let deps = match find_tag_deps dep t  with 
+let rec unprune_dep env r =
+  unprune env r;
+  let deps = match find_tag_deps env.dep r.tag  with 
     | None -> []
     | Some d -> d in
-  List.iter (fun d -> unprune d d.tag dep) deps
+  List.iter (fun d -> unprune_dep env d) deps
 
-let prune_nodep r t =
-  r.pruned <- true;
-  t#set_property (`FOREGROUND "light gray")
+let toggle_incorrect_prune env r =
+  if r.pruned then unprune env r
+  else incorrect_prune env r
 
-let incorrect_prune_nodep r t =
-  r.pruned <- true;
-  t#set_property (`FOREGROUND "tomato")
-
-let unprune_nodep r t =
-  r.pruned <- false;
-  t#set_property (`FOREGROUND_SET false)
-
-let toggle_incorrect_prune r t =
-  if r.pruned then unprune_nodep r t
-  else incorrect_prune_nodep r t
-
-let toggle_prune_nodep r t =
-  if r.pruned then unprune_nodep r t
-  else prune_nodep r t
+let toggle_prune env r =
+  if r.pruned then unprune env r
+  else prune env r
 
 let search_using t sbuf env =
   List.iter (fun t -> t#set_property (`BACKGROUND_SET false)) env.search_tags;
@@ -80,7 +81,26 @@ let search_using t sbuf env =
     
 (* let arrow_cursor () = Gdk.Cursor.create `ARROW *)
 
+let set_select env sbuf = ()
+  (* match env.start_select, env.stop_select with *)
+  (*   | Some b, Some e -> *)
+  (* 	  sbuf#select_range *)
+  (* 	    (sbuf#get_iter (`OFFSET b)) (sbuf#get_iter (`OFFSET e)) *)
+  (*   (\* | None, Some _ -> *\) *)
+  (*   (\* 	if sbuf#has_selection then *\) *)
+  (*   (\* 	  let ib, _ = sbuf#selection_bounds in *\) *)
+  (*   (\* 	  env.start_select <- Some ib#offset; *\) *)
+  (*   (\*       set_select env sbuf *\) *)
+  (*   (\* | Some _, None -> *\) *)
+  (*   (\* 	if sbuf#has_selection then *\) *)
+  (*   (\* 	  let _, ie = sbuf#selection_bounds in *\) *)
+  (*   (\* 	  env.stop_select <- Some ie#offset; *\) *)
+  (*   (\*       set_select env sbuf *\) *)
+	  
+  (*   | _ -> () *)
+
 let tag_callback t env sbuf ~origin:y z i =
+  let ofs = (new GText.iter i)#offset in
   match GdkEvent.get_type z with
     | `MOTION_NOTIFY ->
         if List.mem env.last_tag env.search_tags then 
@@ -105,6 +125,8 @@ let tag_callback t env sbuf ~origin:y z i =
 	    t#set_property (`BACKGROUND "light blue")
 	  end;                         
 	env.last_tag <- t;
+	env.stop_select <- Some ofs;
+	set_select env sbuf;
 	true
     | `TWO_BUTTON_PRESS ->
 	begin
@@ -113,24 +135,25 @@ let tag_callback t env sbuf ~origin:y z i =
 	    | Some an -> match an with
 		| AD (r,_) ->
 		  if env.ctrl then 
-		    if r.pruned then unprune r t env.dep 
-		    else prune r t env.dep
-		  else toggle_prune_nodep r t
+		    if r.pruned then unprune_dep env r 
+		    else prune_dep env r
+		  else toggle_prune env r
 		| AF (r, Some parent) ->
 		  begin match parent.c, parent.polarity with
 		    | AFop (AOPand, _), false | AFop (AOPor, _), true
 		    | AFop (AOPimp, _), true | AFop (AOPiff, _), _ ->
-		        toggle_incorrect_prune r t
-		    | _ -> toggle_prune_nodep r t
+		        toggle_incorrect_prune env r
+		    | _ -> toggle_prune env r
 		  end
-		| AF (r, None) -> toggle_prune_nodep r t
-		| AT r -> toggle_prune_nodep r t
+		| AF (r, None) -> toggle_prune env r
+		| AT r -> toggle_prune env r
 		| QF _ -> ()
 	end;
 	true
     | `BUTTON_PRESS ->
 	let z = GdkEvent.Button.cast z in
-	if GdkEvent.Button.button z = 1 then
+	let captured = 
+	  if GdkEvent.Button.button z = 1 then
 	  if env.ctrl then
 	    (search_using t sbuf env;
 	     true)
@@ -138,7 +161,7 @@ let tag_callback t env sbuf ~origin:y z i =
 	  begin
 	    let tyt = match find t sbuf env.ast with
 	      | Some (AT at) ->
-		  fprintf str_formatter ": %a" Ty.print at.c.at_ty;
+		  fprintf str_formatter ": %a" Ty.print_full at.c.at_ty;
 		  flush_str_formatter ()
 	      | Some (AF _) -> ": formula"
 	      | Some (QF _) -> ": quantified formula"
@@ -153,12 +176,23 @@ let tag_callback t env sbuf ~origin:y z i =
 	    ignore(env.st_ctx#push tyt);
 	    true
 	  end
-	else false
+	  else false
+	in
+	env.start_select <- Some ofs;
+	set_select env sbuf;
+	captured
+
+    | `BUTTON_RELEASE ->
+	env.start_select <- None;
+	env.stop_select <- None;
+	set_select env sbuf;
+	false
+	
     | _ -> false
 
 
 let term_callback t env sbuf ~origin:y z i =
-  if tag_callback t env sbuf ~origin:y z i = true then true
+  if tag_callback t env sbuf ~origin:y z i then true
   else
     match GdkEvent.get_type z with
       | `BUTTON_PRESS ->
@@ -222,7 +256,7 @@ let rec is_quantified_term vars at =
   | ATinfix (at1, _, at2) ->
       is_quantified_term vars at1
       || is_quantified_term vars at2
-  | ATdot (at, _) | ATprefix (_, at) ->
+  | ATdot (at, _) | ATprefix (_, at) | ATnamed (_, at) ->
       is_quantified_term vars at
   | ATextract (at1, at2, at3)
   | ATset (at1, at2, at3) ->
@@ -261,7 +295,7 @@ let rec aterm_used_vars goal_vars at =
 	 with Not_found ->  [])
     | ATapp (_, atl) ->
 	List.fold_left (fun l at -> aterm_used_vars goal_vars at @ l) [] atl
-    | ATdot (at, _) | ATprefix (_, at) | ATlet (_, _, at) ->
+    | ATdot (at, _) | ATprefix (_, at) | ATlet (_, _, at) | ATnamed (_, at) ->
 	aterm_used_vars goal_vars at
     | ATinfix (at1, _, at2) | ATget (at1, at2) | ATconcat (at1, at2) ->
 	(aterm_used_vars goal_vars at1)@(aterm_used_vars goal_vars at2)
@@ -299,10 +333,10 @@ let rec unquantify_aform (buffer:sbuffer) tyenv vars_entries
 	  let ((s, _) as v'), e = List.hd ve in
 	  let cdr_ve = List.tl ve in
 	  assert (v = v');
-	  if e#text = "" then 
+	  if e = "" then 
 	    (v'::nbv, used, goal_used, cdr_ve, v'::uplet, lets)
 	  else
-	    let lb = Lexing.from_string e#text in
+	    let lb = Lexing.from_string e in
 	    let lexpr = Why_parser.lexpr Why_lexer.token lb in
 	    let at, gu =
 	      try 
@@ -387,11 +421,10 @@ let rec unquantify_aform (buffer:sbuffer) tyenv vars_entries
   new_annot buffer c (Why_typing.new_id ()) ptag, ve, goal_used
 
 
-let make_instance (buffer:sbuffer) vars (entries:GEdit.entry list)
-    afc goal_form tyenv =
+let make_instance (buffer:sbuffer) vars entries afc goal_form tyenv =
   let goal_vars = list_vars_in_form goal_form.c in
-  if debug then List.iter (fun (v,e) ->
-    eprintf "%a -> %s@." Symbols.print (fst v) e#text)
+  if debug () then List.iter (fun (v,e) ->
+    eprintf "%a -> %s@." Symbols.print (fst v) e)
     (List.combine vars (List.rev entries));
   let aform, _, goal_used = 
     unquantify_aform buffer tyenv (List.combine vars (List.rev entries)) [] 
@@ -436,7 +469,7 @@ let rec least_nested_form used_vars af =
     | _, AFnamed (_, af) -> 
 	least_nested_form used_vars af
 
-let rec add_instance env vars entries af aname =
+let rec add_instance_aux ?(register=true) env id af aname inv vars entries =
   let ptag =  (tag env.inst_buffer) in
   let goal_form, tyenv, loc =
     let rec find_goal = function
@@ -445,7 +478,7 @@ let rec add_instance env vars entries af aname =
       | x::r -> find_goal r in
     let g, tyenv = find_goal env.ast in
     match g.c with
-      | AGoal (loc, _, f) -> f, tyenv, loc
+      | AGoal (loc, _, _, f) -> f, tyenv, loc
       | _ -> raise Not_found
   in
   let instance, used_vars =
@@ -453,7 +486,8 @@ let rec add_instance env vars entries af aname =
   let ln_form = least_nested_form used_vars goal_form in
   env.inst_buffer#place_cursor  ~where:env.inst_buffer#end_iter;
   if ln_form = Exists goal_form then begin
-    let hy = AAxiom (loc, (sprintf "%s%s" "_instance_" aname), instance.c) in
+    let hy = 
+      AAxiom (loc, (sprintf "%s%s" "_instance_" aname), inv, instance.c) in
     let ahy = new_annot env.inst_buffer hy instance.id ptag in
     let rev_ast = List.rev env.ast in
     let rev_ast = match rev_ast with 
@@ -463,38 +497,49 @@ let rec add_instance env vars entries af aname =
     env.ast <- List.rev rev_ast;
     connect_tag env env.inst_buffer ahy.tag;
     connect_aaform env env.inst_buffer instance;
-    add_to_buffer env.inst_buffer [ahy, tyenv] 
+    add_to_buffer env.errors env.inst_buffer [ahy, tyenv] 
   end
   else begin
     let instance = new_annot env.inst_buffer instance.c instance.id ptag in
     begin match ln_form with 
       | Exists lnf ->
-	lnf.c <- 
-	  AFop 
-	  (AOPand, 
-	   [instance; {lnf with c = lnf.c}])
+	  lnf.c <- 
+	    AFop 
+	    (AOPand, 
+	     [instance; {lnf with c = lnf.c}])
       | Forall lnf ->
-	lnf.c <- 
-	  AFop 
-	  (AOPimp, 
-	   [instance; {lnf with c = lnf.c}])
+	  lnf.c <- 
+	    AFop 
+	    (AOPimp, 
+	     [instance; {lnf with c = lnf.c}])
     end;
     env.inst_buffer#insert ~tags:[instance.tag] ("instance "^aname^": \n");
     connect_aaform env env.inst_buffer instance;
     env.inst_buffer#insert (String.make indent_size ' ');
-    add_aaform env.inst_buffer 1 [] instance;
+    add_aaform env.errors env.inst_buffer 1 [] instance;
     env.inst_buffer#insert "\n\n";
-  end
-    
-  
+  end;
+  if register then save env.actions (AddInstance (id, aname, entries))
+
+
+and add_instance_entries ?(register=true) env id af aname inv 
+    vars (entries:GEdit.entry list) =
+
+  let entries = List.map (fun e -> e#text) entries in
+  add_instance_aux ~register env id af aname inv vars entries
+
+and add_instance ?(register=true) env id af aname inv entries =
+  add_instance_aux ~register env id af aname inv
+    (list_uquant_vars_in_form af) entries
+
 
 and popup_axiom t env offset () =
-    let pop_w = GWindow.dialog
+  let pop_w = GWindow.dialog
     ~title:"Instantiate axiom"
     ~allow_grow:true
     ~width:400 ()
     (* ~icon:(GdkPixbuf.from_xpm_data Logo.xpm_logo) ()  *)
-    in
+  in
   let bbox = GPack.button_box `HORIZONTAL ~border_width:5 ~layout:`END
     ~child_height:20 ~child_width:85 ~spacing:10
     ~packing:pop_w#action_area#add () in
@@ -508,46 +553,51 @@ and popup_axiom t env offset () =
   let phbox = GPack.hbox ~packing:button_cancel#add () in
   ignore(GMisc.image ~stock:`CANCEL ~packing:phbox#add ());
   ignore(GMisc.label ~text:"Cancel" ~packing:phbox#add ());
-  
-  let vars, entries, af, aname = (match find t env.buffer env.ast with
-    | Some (AD (atd, tyenv)) -> 
-      begin
-	  match atd.c with
-	    | AAxiom (_, aname, af) ->
-		pop_w#set_title ("Instantiate axiom "^aname)
-	    | APredicate_def (_, aname,_ , af) ->
-		pop_w#set_title ("Instantiate predicate "^aname)
-	    | _ -> assert false
-      end;
-      begin
-	  match atd.c with
-	    | AAxiom (_, aname, af)
-	    | APredicate_def (_, aname,_ , af) ->
-		let vars = list_uquant_vars_in_form af in
-		let rows = List.length vars in
-		let table = GPack.table ~rows ~columns:2 ~homogeneous:false
-		  ~border_width:5 ~packing:pop_w#vbox#add () in
-		let entries,_ = List.fold_left
-		  (fun (entries,i) (s,ty) ->
-		     fprintf str_formatter "%a : %a = " 
-		       Symbols.print s Ty.print ty;
-		     let text = flush_str_formatter () in
-		     ignore(
-		       GMisc.label ~text ~xalign:1.0
-			 ~packing:(table#attach ~left:0 ~top:i) ());
-		     let entries =
-		       (GEdit.entry ~text:""
-			  ~packing:(table#attach ~left:1 ~top:i
-				      ~expand:`BOTH ~shrink:`BOTH) ()
-		       )::entries in
-		     entries, i+1
-		  ) ([],0) vars in
-		vars, entries, af, aname
-	    | _ -> assert false
-	end
-    | _ -> assert false)
+
+  let vars, entries, id, af, aname, inv = 
+    match find t env.buffer env.ast with
+      | Some (AD (atd, tyenv)) -> 
+	  begin
+	    match atd.c with
+	      | AAxiom (_, aname, _, af) ->
+		  pop_w#set_title ("Instantiate axiom "^aname)
+	      | APredicate_def (_, aname,_ , af) ->
+		  pop_w#set_title ("Instantiate predicate "^aname)
+	      | _ -> assert false
+	  end;
+	  begin
+	    match atd.c with
+	      | AAxiom (_, aname, _, af)
+	      | APredicate_def (_, aname,_ , af) ->
+		  let vars = list_uquant_vars_in_form af in
+		  let rows = List.length vars in
+		  let table = GPack.table ~rows ~columns:2 ~homogeneous:false
+		    ~border_width:5 ~packing:pop_w#vbox#add () in
+		  let entries,_ = List.fold_left
+		    (fun (entries,i) (s,ty) ->
+		       fprintf str_formatter "%a : %a = " 
+			 Symbols.print s Ty.print ty;
+		       let text = flush_str_formatter () in
+		       ignore(
+			 GMisc.label ~text ~xalign:1.0
+			   ~packing:(table#attach ~left:0 ~top:i) ());
+		       let entries =
+			 (GEdit.entry ~text:""
+			    ~packing:(table#attach ~left:1 ~top:i
+					~expand:`BOTH ~shrink:`BOTH) ()
+			 )::entries in
+		       entries, i+1
+		    ) ([],0) vars in
+		  let inv = 
+		    match atd.c with 
+		      | AAxiom(_, _, inv, _ ) -> inv | _ -> false
+		  in
+		  vars, entries, atd.id, af, aname, inv
+	      | _ -> assert false
+	  end
+      | _ -> assert false
   in
-    		
+
   let errors_l = GMisc.label ~text:"" ~packing:pop_w#vbox#pack () in
   errors_l#misc#modify_fg [`NORMAL, `NAME "red"];
   errors_l#misc#hide ();
@@ -555,7 +605,7 @@ and popup_axiom t env offset () =
   ignore(button_ok#connect#clicked ~callback:
 	   (fun () ->
 	      try
-		add_instance env vars entries af aname;
+		add_instance_entries env id af aname inv vars entries;
 		pop_w#destroy ()
 		  
 	      with 
@@ -578,7 +628,7 @@ and popup_axiom t env offset () =
 and axiom_callback t env ~origin:y z i =
   let ni = new GText.iter i in
   let offset = ni#offset in
-  if tag_callback t env env.buffer ~origin:y z i = true then true
+  if tag_callback t env env.buffer ~origin:y z i then true
   else
     begin
       match GdkEvent.get_type z with
@@ -603,15 +653,56 @@ and axiom_callback t env ~origin:y z i =
 
 
 
+and add_trigger ?(register=true) t qid env str offset (sbuf:sbuffer) =
+  let iter = sbuf#get_iter (`OFFSET offset) in
+  match sbuf#forward_iter_to_source_mark 
+    ~category:(sprintf "trigger_%d" qid) iter with
+      | true ->
+	  begin
+	    match find_decl t sbuf env.ast, find t sbuf env.ast with
+	      | Some (AD (_, tyenv)), Some (QF qf) ->
+		  let tags = iter#tags in
+		  let iter = sbuf#get_iter (`OFFSET (iter#offset - 2)) in
+		  let lb = Lexing.from_string str in
+		  let lexprs = Why_parser.trigger Why_lexer.token lb in
+		  let atl = List.fold_right
+		    (fun lexpr l->
+		       let tt = Why_typing.term tyenv
+			 (qf.c.aqf_upvars@qf.c.aqf_bvars) lexpr in
+		       let at = annot_of_tterm sbuf tt in
+		       at.tag#set_priority (t#priority - 1);
+		       connect_aaterm env sbuf connect_tag at;
+		       at::l
+		    ) lexprs [] in
+		  if qf.c.aqf_triggers <> [] then
+		    sbuf#insert ~iter ~tags " | ";
+		  add_aaterm_list_at sbuf tags iter "," atl;
+		  qf.c.aqf_triggers <- qf.c.aqf_triggers@[atl];
+		  if register then 
+		    save env.actions (AddTrigger (qf.id, sbuf=env.inst_buffer, str));
+	      | _ -> assert false
+	  end
+      | false -> ()
+
+and readd_trigger ?(register=true) env id str inst_buf =
+  try
+    match findbyid id env.ast with
+      | QF qf ->
+	  let sbuf = 
+	    if inst_buf then env.inst_buffer else env.buffer in
+	  add_trigger ~register qf.tag id env str 0 sbuf
+      | _ -> assert false
+  with Not_found -> ()
 
 
-and popup_trigger t env (sbuf:sbuffer) offset () =
-    let pop_w = GWindow.dialog
+and popup_trigger t qid env (sbuf:sbuffer) offset () =
+  let pop_w = GWindow.dialog
     ~title:"Add new (multi) trigger"
     ~allow_grow:true
-    ~width:400 ()
+    ~width:400 
+    ~height:100 ()
     (* ~icon:(GdkPixbuf.from_xpm_data Logo.xpm_logo) ()  *)
-    in
+  in
   let bbox = GPack.button_box `HORIZONTAL ~border_width:5 ~layout:`END
     ~child_height:20 ~child_width:85 ~spacing:10
     ~packing:pop_w#action_area#add () in
@@ -637,66 +728,37 @@ and popup_trigger t env (sbuf:sbuffer) offset () =
     ~hpolicy:`AUTOMATIC
     ~packing:pop_w#vbox#add () in
   let tv1 = GSourceView2.source_view ~source_buffer:buf1 ~packing:(sw1#add) 
-	 ~show_line_numbers:true ~wrap_mode:`CHAR() in
+    ~show_line_numbers:true ~wrap_mode:`CHAR() in
   let _ = tv1#misc#modify_font monospace_font in
   let _ = tv1#set_editable true in
-		
+  
   let errors_l = GMisc.label ~text:"" ~packing:pop_w#vbox#pack () in
   errors_l#misc#modify_fg [`NORMAL, `NAME "red"];
   errors_l#misc#hide ();
-    
+  
   ignore(button_ok#connect#clicked
-    ~callback: 
-    (fun () ->
-       let iter = sbuf#get_iter (`OFFSET offset) in
-       try begin
-	 match sbuf#forward_iter_to_source_mark ~category:"trigger" iter with
-	   | true ->
-	       begin
-		 match find_decl t sbuf env.ast, find t sbuf env.ast with
-		   | Some (AD (_, tyenv)), Some (QF qf) ->
-		       
-		       (* let iter = sbuf#get_iter_at_marker m in *)
-		       let tags = iter#tags in
-		       let iter = sbuf#get_iter
-			 (`OFFSET (iter#offset - 2)) in
-		       let str = buf1#get_text () in
-		       
-		       let lb = Lexing.from_string str in
-		       let lexprs = Why_parser.trigger Why_lexer.token lb in
-		       let atl = List.fold_right
-			 (fun lexpr l->
-			    let tt = Why_typing.term tyenv
-			      (qf.c.aqf_upvars@qf.c.aqf_bvars) lexpr in
-			    let at = annot_of_tterm sbuf tt in
-			    at.tag#set_priority (t#priority - 1);
-			    connect_aaterm env sbuf connect_tag at;
-			    at::l
-			 ) lexprs [] in		       
-		       sbuf#insert ~iter ~tags " | ";
-		       add_aaterm_list_at sbuf tags iter "," atl;
-		       qf.c.aqf_triggers <- qf.c.aqf_triggers@[atl];
-		   | _ -> assert false
-	       end
-	   | false -> ()
-       end;
-	 pop_w#destroy ()
-       with 
-    | Why_lexer.Lexical_error s -> 
-	errors_l#set_text ("Lexical error");
-	errors_l#misc#show ()
-    | Parsing.Parse_error ->
-	errors_l#set_text ("Syntax error");
-	errors_l#misc#show ()
-    | Common.Error (e,_) ->
-	fprintf str_formatter "Typing error : %a" Common.report e;
-	errors_l#set_text (flush_str_formatter ());
-	errors_l#misc#show ()
-    ));
+	   ~callback: 
+	   (fun () ->
+	      try 
+		let str = buf1#get_text () in
+		add_trigger t qid env str offset sbuf;
+		pop_w#destroy ()
+	      with 
+		| Why_lexer.Lexical_error s -> 
+		    errors_l#set_text ("Lexical error");
+		    errors_l#misc#show ()
+		| Parsing.Parse_error ->
+		    errors_l#set_text ("Syntax error");
+		    errors_l#misc#show ()
+		| Common.Error (e,_) ->
+		    fprintf str_formatter "Typing error : %a" Common.report e;
+		    errors_l#set_text (flush_str_formatter ());
+		    errors_l#misc#show ()
+	   ));
   ignore(button_cancel#connect#clicked ~callback: pop_w#destroy);
   pop_w#show ()
 
-and triggers_callback t env sbuf ~origin:y z i =
+and triggers_callback t qid env sbuf ~origin:y z i =
   
   let ni = new GText.iter i in
   let offset = ni#offset in
@@ -712,7 +774,7 @@ and triggers_callback t env sbuf ~origin:y z i =
 	      let menuitem = GMenu.image_menu_item ~image
 		~label:"Add trigger(s) ..." ~packing:menu#append () in
 	      ignore(menuitem#connect#activate
-		       ~callback:(popup_trigger t env sbuf offset));
+		       ~callback:(popup_trigger t qid env sbuf offset));
 	      menu#popup ~button:3 ~time:(GdkEvent.Button.time z);
 	      true
 	    else
@@ -726,7 +788,7 @@ and triggers_callback t env sbuf ~origin:y z i =
 (*   ignore (t#connect#event ~callback:(set_mark t buffer)); *)
 (*   (\* ignore (t#connect#event ~callback:(fetch_text t buffer)); *\) *)
 (*   t *)
-  
+      
 
 and connect_tag env sbuf t =
   ignore (t#connect#event ~callback:(tag_callback t env sbuf))
@@ -734,8 +796,8 @@ and connect_tag env sbuf t =
 and connect_term env sbuf t =
   ignore (t#connect#event ~callback:(term_callback t env sbuf))
 
-and connect_trigger_tag env sbuf t =
-  ignore (t#connect#event ~callback:(triggers_callback t env sbuf))
+and connect_trigger_tag env sbuf t qid =
+  ignore (t#connect#event ~callback:(triggers_callback t qid env sbuf))
 
 and connect_axiom_tag env t =
   ignore (t#connect#event ~callback:(axiom_callback t env))
@@ -756,20 +818,21 @@ and connect_aaterm_list env sbuf
   List.iter (connect_aaterm env sbuf connect_tag) aatl
 
 and connect_at_desc env sbuf = function
-    | ATconst _ | ATvar _ -> ()
-    | ATapp (s, atl) -> connect_aterm_list env sbuf atl
-    | ATinfix (t1, _, t2) | ATget (t1, t2)
-    | ATconcat (t1, t2) | ATlet (_, t1, t2) ->
-	connect_aterm env sbuf t1;
-	connect_aterm env sbuf t2
-    | ATdot (t, _) | ATprefix (_, t) -> connect_aterm env sbuf t
-    | ATset (t1,t2,t3) | ATextract (t1,t2,t3) ->
-	connect_aterm env sbuf t1;
-	connect_aterm env sbuf t2;
-	connect_aterm env sbuf t3
-    | ATrecord r -> 
-	let atl = List.map snd r in
-	connect_aterm_list env sbuf atl
+  | ATconst _ | ATvar _ -> ()
+  | ATapp (s, atl) -> connect_aterm_list env sbuf atl
+  | ATinfix (t1, _, t2) | ATget (t1, t2)
+  | ATconcat (t1, t2) | ATlet (_, t1, t2) ->
+      connect_aterm env sbuf t1;
+      connect_aterm env sbuf t2
+  | ATdot (t, _) | ATprefix (_, t) | ATnamed (_, t) ->
+      connect_aterm env sbuf t
+  | ATset (t1,t2,t3) | ATextract (t1,t2,t3) ->
+      connect_aterm env sbuf t1;
+      connect_aterm env sbuf t2;
+      connect_aterm env sbuf t3
+  | ATrecord r -> 
+      let atl = List.map snd r in
+      connect_aterm_list env sbuf atl
 	
 and connect_aatom env sbuf aa =
   match aa with
@@ -792,13 +855,13 @@ and connect_quant_form env sbuf
 
 and connect_triggers env sbuf trs =
   List.iter (connect_aaterm_list env sbuf connect_tag) trs
-      
+    
 and connect_aform env sbuf = function
   | AFatom a -> connect_aatom env sbuf a
   | AFop (op, afl) -> connect_aaform_list env sbuf afl
   | AFforall aqf
   | AFexists aqf -> 
-      connect_trigger_tag env sbuf aqf.tag;
+      connect_trigger_tag env sbuf aqf.tag aqf.id;
       connect_quant_form env sbuf aqf.c
   | AFlet (vs, s, t, aaf) ->
       connect_aterm env sbuf t;      
@@ -816,13 +879,13 @@ and connect_aaform env sbuf aaf =
 let connect_atyped_decl env td =
   match td.c with
     | APredicate_def (_, _, _, af)
-    | AAxiom (_, _, af) ->
+    | AAxiom (_, _, _, af) ->
 	connect_axiom_tag env td.tag;
 	connect_aform env env.buffer af
     | ARewriting (_, _, arwtl) ->
 	connect_tag env env.buffer td.tag
 	(* TODO *)
-    | AGoal (_, _, aaf) ->
+    | AGoal (_, _, _, aaf) ->
 	connect_tag env env.buffer td.tag;
 	connect_aform env env.buffer aaf.c
     | AFunction_def (_, _, _, _, af) ->
@@ -836,17 +899,23 @@ let connect env =
   List.iter (fun (t, _) -> connect_atyped_decl env t) env.ast
 
 let clear_used_lemmas_tags env =
-  List.iter (fun t -> t#set_property (`BACKGROUND_SET false)) env.proof_tags;
+  MTag.iter (fun t _ -> t#set_property (`BACKGROUND_SET false)) env.proof_tags;
   List.iter (fun t -> t#set_property (`BACKGROUND_SET false)) env.proof_toptags;
-  env.proof_tags <- [];
+  env.proof_tags <- MTag.empty;
   env.proof_toptags <- []
   
-
 let show_used_lemmas env expl =
+  let colormap = Gdk.Color.get_system_colormap () in
   let atags,ftags = findtags_proof expl env.ast in
   clear_used_lemmas_tags env;
-  List.iter (fun t -> t#set_property (`BACKGROUND "pale green")) atags;
-  List.iter (fun t -> t#set_property (`BACKGROUND "green")) ftags;
+  let max_mul = MTag.fold (fun _ m acc -> max acc m) ftags 0 in
+  let green_0 = Gdk.Color.alloc colormap (`RGB (65535*3/4, 65535, 65535*3/4)) in
+  List.iter (fun t -> t#set_property (`BACKGROUND_GDK green_0)) atags;
+  MTag.iter (fun t m ->
+    let perc = ((max_mul - m) * 65535) / max_mul in
+    let green_n = Gdk.Color.alloc colormap 
+      (`RGB (perc*1/2, (perc + 2*65535) /3, perc*1/2)) in
+    t#set_property (`BACKGROUND_GDK green_n)) ftags;
   env.proof_tags <- ftags;
   env.proof_toptags <- atags
   
@@ -859,7 +928,7 @@ let prune_unused env expl =
   in
   let prune_top d = match d.c with
     | ATypeDecl _ | AGoal _ | ALogic _ -> ()
-    | _ -> prune_nodep d d.tag
+    | _ -> prune d d.tag
   in
   let rec aux dont ast ids =
     match ast, ids with
@@ -882,10 +951,10 @@ let prune_unused env expl =
 let prune_unused env =
   let prune_top d = match d.c with
     | ATypeDecl _ | AGoal _ | ALogic _ -> ()
-    | _ -> prune_nodep d d.tag
+    | _ -> prune env d
   in
   List.iter (fun (d, _) -> 
     if not (List.mem d.ptag env.proof_toptags) 
-      && not (List.mem d.ptag env.proof_tags) 
+      && not (MTag.mem d.ptag env.proof_tags) 
     then prune_top d
   ) env.ast
